/*
Copyright (C) 2001 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"

#include "D3DRender.h"
#include "DLParser.h"

#include "Utility\Profiler.h"
#include "Debug\DBGConsole.h"

#include "Core\Memory.h"			// We access the memory buffers
#include "PngUtil.h"

#include "RDP.h"

#define MAX_TEXTURE_STAGES 8

#ifdef USE_PIXEL_SHADERS
	DWORD PS_Combiner = 0;
#else
	bool	gUsingPrimColour = false;
	bool	gUsingEnvColour = false;
#endif

//*****************************************************************************
// Creator function for singleton
//*****************************************************************************
HRESULT CUnique< D3DRender >::Create()
{
	_ASSERTE(m_pInstance == NULL);
	
	m_pInstance = new D3DRender();
	if (!m_pInstance)
	{
		return E_OUTOFMEMORY;
	}
	
	return S_OK;	
}


//*****************************************************************************
//
//*****************************************************************************
class CombineStage 
{
public:
	DWORD dwColorOp, dwColorArg1, dwColorArg2;
	DWORD dwAlphaOp, dwAlphaArg1, dwAlphaArg2;

	BOOL operator!=(const CombineStage & cs) const { return !(operator==(cs)); }
	BOOL operator==(const CombineStage & cs) const 
	{
		return (
			cs.dwAlphaArg1 == dwAlphaArg1 &&
			cs.dwAlphaArg2 == dwAlphaArg2 &&
			cs.dwAlphaOp == dwAlphaOp &&
			cs.dwColorArg1 == dwColorArg1 &&
			cs.dwColorArg2 == dwColorArg2 &&
			cs.dwColorOp == dwColorOp);

	}

};	


enum MuxVal
{
	MUX_0 = 0,
	MUX_1,
	MUX_COMBINED,
	MUX_TEXEL0,
	MUX_TEXEL1,
	MUX_PRIM,
	MUX_SHADE,
	MUX_ENV,
	MUX_COMBALPHA,
	MUX_T0_ALPHA,
	MUX_T1_ALPHA,
	MUX_PRIM_ALPHA,
	MUX_SHADE_ALPHA,
	MUX_ENV_ALPHA,
	MUX_LODFRAC,
	MUX_PRIMLODFRAC,
	MUX_UNK,
	MUX_PRIM_MOD_ENV,
	MUX_PRIM_ENV_MEAN
};

//*****************************************************************************
//
//*****************************************************************************
class SetCombineInfo
{
public:
	u64 mux;
	LONG nStages;

	BOOL  bAlphaEnabled;

	// Added by Orkin - tells InitBlendMode() what to set the texture factor to
	MuxVal TFactor;

	CombineStage stages[MAX_TEXTURE_STAGES];

	BOOL operator!=(const SetCombineInfo & sci) const { return !(operator==(sci)); }
	BOOL operator==(const SetCombineInfo & sci) const 
	{
		LONG i;

		if (sci.nStages != nStages)
			return FALSE;
		if (sci.bAlphaEnabled != bAlphaEnabled)
			return FALSE;

		for (i = 0; i < nStages; i++)
		{
			if (sci.stages[i] != stages[i])
				return FALSE;
		}
		return TRUE;
	}
};



static std::vector<u64> g_UnkMux;
static std::vector<u64> g_FoundMux;
static std::vector<u64> g_MismatchedMux;
static std::vector<SetCombineInfo> g_GeneratedMux;



#define D3DTA_IGNORE ~0

// I moved these to separate files to reduce the size of this module
// (the combiner table is getting huge now!)
#include "D3DRender_Combine.inl"


//*****************************************************************************
//
//*****************************************************************************
D3DRender::D3DRender() :
#ifdef VRML_EXPORT_SUPPORT
	m_DumpingScene( false ),
#endif
	mViWidth( 320 ),
	mViHeight( 240 ),
	m_fScreenMultX(2.0f),
	m_fScreenMultY(2.0f),

	m_dwProjectionTop(0),
	m_dwModelViewTop(0),
		
	m_dwTextureFactor(D3DCOLOR_RGBA(255,255,255,255)),

	m_bCullFront(FALSE),
	m_bCullBack(TRUE),

	m_bTextureEnable(FALSE),
	m_bLightingEnable(FALSE),
	m_dwNumLights(0),
	
	m_bZBuffer(FALSE),
	
	m_dwAlpha(0xFF),

	mSmooth( true ),
	mSmoothShade( true ),

	m_bTextureGen(FALSE),

	m_dwNumVertices(0),
	m_dwNumIndices(0),
	m_lpVtxBuffer( NULL ),
	m_lpIdxBuffer( NULL ),

	m_fFogMult(0.0f),			// * 0 + 0 = 0
	m_fFogOffset(0.0f),
	m_bFogEnable(FALSE),
	m_dwFogColor(0x00FFFFFF),

	m_dwNumTrisRendered( 0 ),
	m_dwNumTrisClipped( 0 )
{
	for ( u32 t = 0; t < 2; t++ )
	{
		m_lpsTexture[t] = NULL;
		m_fTileWidth[t] = 32.0f;
		m_fTileHeight[t] = 32.0f;
		m_fTexWidth[t] = 32.0f;
		m_fTexHeight[t] = 32.0f;
		m_fInvTileWidth[t] = 1/32.0f;
		m_fInvTileHeight[t] = 1/32.0f;
		m_fInvTexWidth[t] = 1/32.0f;
		m_fInvTexHeight[t] = 1/32.0f;
	}
	m_fTexScaleX = 1.0f;
	m_fTexScaleY = 1.0f;

	m_AmbientCol.r = m_AmbientCol.g = m_AmbientCol.b = m_AmbientCol.a = 1.0f;

	gRDPMux._u64 = 0;

	// Perform a quick test to ensure the table is correctly ordered
	u64 muxlast = g_SCInfo[0].mux;
	u64 mux;

	u32 i;
	for ( i = 1; i < NUM_SCINFO; i++ )
	{
		mux = g_SCInfo[i].mux;
		if (mux <= muxlast)
		{
			CHAR str[300];
			wsprintf(str, "Warning, g_SCInfo table is not ordered: %08x%08x",
				(u32)(mux>>32), (u32)mux);
			MessageBox(NULL, str, g_szDaedalusName, MB_OK);
			break;
		}
		muxlast = mux;
	}


}

//*****************************************************************************
//
//*****************************************************************************
D3DRender::~D3DRender()
{	
	if (g_pD3DDev && m_lpVtxBuffer)
	{
		// Necessary? Seems to crash...
		//g_pD3DDev->SetStreamSource(0, NULL, 0);
	}

	InvalidateDeviceObjects();
}
	



//*****************************************************************************
//
//*****************************************************************************
HRESULT D3DRender::RestoreDeviceObjects()
{
	HRESULT hr;

	if ( !g_pD3DDev )
		return E_FAIL;

	// Initialise the device to our default state

	// No longer Needed for mario menu screen
	g_pD3DDev->SetRenderState( D3DRS_DITHERENABLE, TRUE ); // Re-enabled by Orkin - Makes 16-bit look nicer
	//g_pD3DDev->SetRenderState( D3DRS_SPECULARENABLE, FALSE );

	// We do our own culling
	g_pD3DDev->SetRenderState( D3DRS_CULLMODE,   D3DCULL_NONE );

	// We do our own lighting
	//g_pD3DDev->SetRenderState( D3DRS_AMBIENT, D3DCOLOR_XRGB(255,255,255) );
	g_pD3DDev->SetRenderState( D3DRS_LIGHTING,	  FALSE);

	g_pD3DDev->SetRenderState( D3DRS_ALPHABLENDENABLE, TRUE );
	g_pD3DDev->SetRenderState( D3DRS_SRCBLEND,   D3DBLEND_SRCALPHA);
	g_pD3DDev->SetRenderState( D3DRS_DESTBLEND,  D3DBLEND_INVSRCALPHA);

	g_pD3DDev->SetRenderState( D3DRS_ZFUNC, D3DCMP_LESSEQUAL);
	
	// Initialise all the renderstate to our defaults.
	g_pD3DDev->SetRenderState( D3DRS_SHADEMODE,  D3DSHADE_GOURAUD );
	g_pD3DDev->SetRenderState( D3DRS_TEXTUREFACTOR, m_dwTextureFactor );

	g_pD3DDev->SetRenderState( D3DRS_FOGCOLOR, m_dwFogColor ); 
	//g_pD3DDev->SetRenderState (D3DRS_FOGVERTEXMODE, D3DFOG_LINEAR );
        
	// Dafault is ZBuffer disabled
	g_pD3DDev->SetRenderState( D3DRS_ZENABLE,      D3DZB_FALSE );
	g_pD3DDev->SetRenderState( D3DRS_ZWRITEENABLE, FALSE );

	//g_pD3DDev->SetRenderState( D3DRS_TEXTUREPERSPECTIVE, TRUE );
	
	//g_pD3DDev->SetRenderState( D3DRS_ALPHATESTENABLE, m_bAlphaTestEnable );
	//g_pD3DDev->SetRenderState(D3DRS_ALPHAREF, m_dwAlpha);
    g_pD3DDev->SetRenderState( D3DRS_ALPHATESTENABLE, TRUE );
    g_pD3DDev->SetRenderState( D3DRS_ALPHAREF,        0x04 );
    g_pD3DDev->SetRenderState( D3DRS_ALPHAFUNC, D3DCMP_GREATEREQUAL );

	// Texturing stuff
	// Thanks Lkb
	for(int n = 0; n < MAX_TEXTURE_STAGES; n++) 
	{ 
		// Texturing stuff 
		g_pD3DDev->SetTextureStageState( n, D3DTSS_MINFILTER, D3DTEXF_LINEAR ); 
		g_pD3DDev->SetTextureStageState( n, D3DTSS_MAGFILTER, D3DTEXF_LINEAR ); 
		
		g_pD3DDev->SetTextureStageState( n, D3DTSS_ADDRESSU, D3DTADDRESS_WRAP ); 
		g_pD3DDev->SetTextureStageState( n, D3DTSS_ADDRESSV, D3DTADDRESS_WRAP ); 
		
		if(n == 0) 
		{ 
			g_pD3DDev->SetTextureStageState( n, D3DTSS_COLORARG1, D3DTA_TEXTURE ); 
			g_pD3DDev->SetTextureStageState( n, D3DTSS_COLORARG2, D3DTA_DIFFUSE ); 
			g_pD3DDev->SetTextureStageState( n, D3DTSS_COLOROP, D3DTOP_MODULATE ); 
			g_pD3DDev->SetTextureStageState( n, D3DTSS_ALPHAARG1, D3DTA_TEXTURE ); 
			g_pD3DDev->SetTextureStageState( n, D3DTSS_ALPHAARG2, D3DTA_DIFFUSE ); 
			g_pD3DDev->SetTextureStageState( n, D3DTSS_ALPHAOP, D3DTOP_MODULATE ); 
		} 
		else 
		{ 
			g_pD3DDev->SetTextureStageState( n, D3DTSS_COLOROP, D3DTOP_DISABLE ); 
			g_pD3DDev->SetTextureStageState( n, D3DTSS_ALPHAOP, D3DTOP_DISABLE ); 
		} 
	} 

	if ( m_lpVtxBuffer )
	{
		m_lpVtxBuffer->Release();
		m_lpVtxBuffer = NULL;
	}

	// Create out vertex buffer
	hr = g_pD3DDev->CreateVertexBuffer( 500*sizeof(DaedalusVtx),		// Size in bytes
									    D3DUSAGE_WRITEONLY|D3DUSAGE_DYNAMIC ,
									    FVF_DaedalusVtx,	// Or 0
									    D3DPOOL_DEFAULT,			// Can't be MANAGED with DYNAMIC usage
									   &m_lpVtxBuffer );
	m_dwVtxBufBase = 500;			// Forces us to start at the beginning of the buffer with a discard
	m_dwVtxBufSize = 500;

	if (FAILED(hr)) 
		return hr;


	if ( m_lpIdxBuffer )
	{
		m_lpIdxBuffer->Release();
		m_lpIdxBuffer = NULL;
	}

	hr = g_pD3DDev->CreateIndexBuffer( 500 * sizeof(WORD),
									   D3DUSAGE_WRITEONLY|D3DUSAGE_DYNAMIC, 
									   D3DFMT_INDEX16,
									   D3DPOOL_DEFAULT,
									   &m_lpIdxBuffer );
	m_dwIdxBufBase = 500;
	m_dwIdxBufSize = 500;
	if (FAILED(hr))
		return hr;

	return S_OK;		
}

//*****************************************************************************
//
//*****************************************************************************
// Reset for a new frame
void	D3DRender::Reset()
{
	ResetMatrices();

	m_dwNumVertices = 0;
	m_dwNumIndices = 0;

	m_dwNumTrisRendered = 0;
	m_dwNumTrisClipped = 0;

	u32 dwScaleX = Memory_VI_GetRegister( VI_X_SCALE_REG ) & 0xFFF;
	u32 dwScaleY = Memory_VI_GetRegister( VI_Y_SCALE_REG ) & 0xFFF;

	f32 fScaleX = f32(dwScaleX) / (1<<10);
	f32 fScaleY = f32(dwScaleY) / (1<<10);

	u32 dwHStartReg = Memory_VI_GetRegister( VI_H_START_REG );
	u32 dwVStartReg = Memory_VI_GetRegister( VI_V_START_REG );

	u32	hstart = dwHStartReg >> 16;
	u32	hend = dwHStartReg & 0xffff;
	//DBGConsole_Msg( 0, "h start/end %d %d", hstart, hend );		// 128 725 - 597

	u32	vstart = dwVStartReg >> 16;
	u32	vend = dwVStartReg & 0xffff;
	//DBGConsole_Msg( 0, "v start/end %d %d", vstart, vend );		// 56 501 - 445

	mViWidth  = (u32)( (hend-hstart)    * fScaleX);
	mViHeight = (u32)(((vend-vstart)/2) * fScaleY);					// /2 is because vertical is measured in half-lines

	if ( mViWidth == 0 )
	{
		mViWidth = 320;
	}
	if ( mViHeight == 0 )
	{
		mViHeight = 240;
	}
}


//*****************************************************************************
//
//*****************************************************************************
void	D3DRender::InitCycleCopy( )
{
	DL_PF("CycleType: Copy");

	g_pD3DDev->SetTextureStageState( 1, D3DTSS_COLOROP,   D3DTOP_DISABLE );
	g_pD3DDev->SetTextureStageState( 1, D3DTSS_ALPHAOP,   D3DTOP_DISABLE );
	
	g_pD3DDev->SetTexture( 0, m_lpsTexture[0] );
	g_pD3DDev->SetTexture( 1, NULL );

	g_pD3DDev->SetTextureStageState( 0, D3DTSS_COLOROP, D3DTOP_SELECTARG1 );
	g_pD3DDev->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TEXTURE );

	g_pD3DDev->SetTextureStageState( 0, D3DTSS_ALPHAOP, D3DTOP_SELECTARG1 );
	g_pD3DDev->SetTextureStageState( 0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE );

	g_pD3DDev->SetPixelShader( NULL );
}

//*****************************************************************************
//
//*****************************************************************************
void	D3DRender::InitCycleFill()
{
	DL_PF("CycleType: Fill");
	
	g_pD3DDev->SetTextureStageState( 1, D3DTSS_COLOROP,   D3DTOP_DISABLE );
	g_pD3DDev->SetTextureStageState( 1, D3DTSS_ALPHAOP,   D3DTOP_DISABLE );

	g_pD3DDev->SetTexture( 0, NULL );
	g_pD3DDev->SetTexture( 1, NULL );

	g_pD3DDev->SetTextureStageState( 0, D3DTSS_COLOROP, D3DTOP_SELECTARG1 );
	g_pD3DDev->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_DIFFUSE );

	g_pD3DDev->SetTextureStageState( 0, D3DTSS_ALPHAOP, D3DTOP_SELECTARG1 );
	g_pD3DDev->SetTextureStageState( 0, D3DTSS_ALPHAARG1, D3DTA_DIFFUSE );

	g_pD3DDev->SetPixelShader( NULL );
}

//*************************************************************************************
// 
//*************************************************************************************
void	D3DRender::InitCycle1()
{
	DL_PF("CycleType: 1");

#ifdef USE_PIXEL_SHADERS

	if ( m_lpsTexture[0] )
	{
		InitCycle1PixelShader();

		g_pD3DDev->SetTexture( 0, m_lpsTexture[0] );
		if ( m_lpsTexture[1] )
		{
			g_pD3DDev->SetTexture( 1, m_lpsTexture[1] );
		}
		else
		{
			g_pD3DDev->SetTexture( 1, m_lpsTexture[0] );
		}
	}
	else
	{
		g_pD3DDev->SetPixelShader( NULL );
	}

#else 

	InitCycle12( false );

#endif
}

//*************************************************************************************
// 
//*************************************************************************************
void	D3DRender::InitCycle2()
{
	DL_PF("CycleType: 2");

#ifdef USE_PIXEL_SHADERS

	if ( m_lpsTexture[0] )
	{
		if ( (gRDPOtherMode._u32[0] & 0xff000000) == 0xc8000000 )
		{
			InitCycle2FogShadePixelShader();
		}
		else
		{
			InitCycle2PixelShader();
		}

		g_pD3DDev->SetTexture( 0, m_lpsTexture[0] );
		if ( m_lpsTexture[1] )
		{
			g_pD3DDev->SetTexture( 1, m_lpsTexture[1] );
		}
		else
		{
			g_pD3DDev->SetTexture( 1, m_lpsTexture[0] );
		}
	}
	else
	{
		g_pD3DDev->SetPixelShader( NULL );
	}

#else 

	InitCycle12( true );

#endif
}

#ifndef USE_PIXEL_SHADERS
//*****************************************************************************
//
//*****************************************************************************
void	D3DRender::InitCycle12( bool two_cycles )
{
	SetCombineInfo sci;

	BOOL bFound = FindCombineMode( gRDPMux._u64, two_cycles, sci );
	
	if (!bFound)
	{
		g_pD3DDev->SetTextureStageState( 1, D3DTSS_COLOROP,   D3DTOP_DISABLE );
		g_pD3DDev->SetTextureStageState( 1, D3DTSS_ALPHAOP,   D3DTOP_DISABLE );
	
		// Set to defaults
		g_pD3DDev->SetTexture( 0, m_lpsTexture[0] );
		g_pD3DDev->SetTexture( 1, m_lpsTexture[1] );

		g_pD3DDev->SetTextureStageState( 0, D3DTSS_COLOROP, D3DTOP_MODULATE );	
		g_pD3DDev->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TEXTURE );
		g_pD3DDev->SetTextureStageState( 0, D3DTSS_COLORARG2, D3DTA_DIFFUSE );

		g_pD3DDev->SetTextureStageState( 0, D3DTSS_ALPHAOP, D3DTOP_MODULATE );
		g_pD3DDev->SetTextureStageState( 0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE );
		g_pD3DDev->SetTextureStageState( 0, D3DTSS_ALPHAARG2, D3DTA_DIFFUSE );
	}
	else // Redone by Orkin - allows for multiple texture stages
	{
		// Allows a combine mode to select what TFACTOR should be
		gUsingPrimColour = false;
		gUsingEnvColour = false;

		switch (sci.TFactor)
		{
			case MUX_0:		SetTextureFactor(D3DCOLOR_RGBA(0, 0, 0, 0));								break;
			case MUX_1:		SetTextureFactor(D3DCOLOR_RGBA(1, 1, 1, 1));								break;
			case MUX_PRIM:	SetTextureFactor(m_dwPrimitiveColor);			gUsingPrimColour = true;	break;
			case MUX_ENV:	SetTextureFactor(m_dwEnvColor);					gUsingEnvColour = true;		break;
			case MUX_PRIM_MOD_ENV:
			{
				float prim_a = ((u8)(m_dwPrimitiveColor >> 24)) / 255.0f;
				float prim_r = ((u8)(m_dwPrimitiveColor >> 16)) / 255.0f;
				float prim_g = ((u8)(m_dwPrimitiveColor >>  8)) / 255.0f;
				float prim_b = ((u8)(m_dwPrimitiveColor      )) / 255.0f;

				float env_a = ((u8)(m_dwEnvColor >> 24)) / 255.0f;
				float env_r = ((u8)(m_dwEnvColor >> 16)) / 255.0f;
				float env_g = ((u8)(m_dwEnvColor >>  8)) / 255.0f;
				float env_b = ((u8)(m_dwEnvColor      )) / 255.0f;

				u32 col = D3DCOLOR_COLORVALUE( prim_r * env_r, prim_g * env_g, prim_b * env_b, prim_a * env_a );
				SetTextureFactor( col );
			}
			break;
			case MUX_PRIM_ENV_MEAN:
			{
				float prim_a = ((u8)(m_dwPrimitiveColor >> 24)) / 255.0f;
				float prim_r = ((u8)(m_dwPrimitiveColor >> 16)) / 255.0f;
				float prim_g = ((u8)(m_dwPrimitiveColor >>  8)) / 255.0f;
				float prim_b = ((u8)(m_dwPrimitiveColor      )) / 255.0f;

				float env_a = ((u8)(m_dwEnvColor >> 24)) / 255.0f;
				float env_r = ((u8)(m_dwEnvColor >> 16)) / 255.0f;
				float env_g = ((u8)(m_dwEnvColor >>  8)) / 255.0f;
				float env_b = ((u8)(m_dwEnvColor      )) / 255.0f;

				u32 col = D3DCOLOR_COLORVALUE( (prim_r + env_r) / 2.0f, (prim_g * env_g) / 2.0f, (prim_b * env_b) / 2.0f, (prim_a * env_a) / 2.0f );
				SetTextureFactor( col );
			}
		}

		for (int i = 0; i < sci.nStages; i++)
		{
			g_pD3DDev->SetTextureStageState( i, D3DTSS_COLOROP,   sci.stages[i].dwColorOp );
			g_pD3DDev->SetTextureStageState( i, D3DTSS_ALPHAOP,   sci.stages[i].dwAlphaOp );

			g_pD3DDev->SetTexture( i, m_lpsTexture[0] );

			g_pD3DDev->SetTextureStageState( i, D3DTSS_COLOROP, sci.stages[i].dwColorOp );

			if ( sci.stages[i].dwColorArg1 != ~0 )		g_pD3DDev->SetTextureStageState( i, D3DTSS_COLORARG1, sci.stages[i].dwColorArg1 );
			if ( sci.stages[i].dwColorArg2 != ~0 )		g_pD3DDev->SetTextureStageState( i, D3DTSS_COLORARG2, sci.stages[i].dwColorArg2 );

			if (sci.bAlphaEnabled)
			{
				g_pD3DDev->SetTextureStageState( i, D3DTSS_ALPHAOP, sci.stages[i].dwAlphaOp );

				if ( sci.stages[i].dwAlphaArg1 != ~0 )		g_pD3DDev->SetTextureStageState( i, D3DTSS_ALPHAARG1, sci.stages[i].dwAlphaArg1 );
				if ( sci.stages[i].dwAlphaArg2 != ~0 )		g_pD3DDev->SetTextureStageState( i, D3DTSS_ALPHAARG2, sci.stages[i].dwAlphaArg2 );
			}
		}
		// Cap off the texture stages
		g_pD3DDev->SetTextureStageState( sci.nStages, D3DTSS_COLOROP,   D3DTOP_DISABLE );
		g_pD3DDev->SetTextureStageState( sci.nStages, D3DTSS_ALPHAOP,   D3DTOP_DISABLE );
	}
}
#endif

//*****************************************************************************
// States set/modified by this mode:
// SetTextureStateState COLOROP/ALPHAOP
// SetTexture
// D3DRS_TEXTUREFACTOR
// D3DRS_ZENABLE
// D3DRS_ZWRITEENABLE
// ---D3DRS_TEXTUREPERSPECTIVE 
// D3DRS_ALPHATESTENABLE
// D3DRS_SRCBLEND
// D3DRS_DESTBLEND
// D3DRS_ALPHABLENDENABLE
//*****************************************************************************
RDP_OtherMode	gLastRDPOtherMode;
RDP_Mux			gLastRDPMux;
bool			gTextureChanged = false;
DWORD			gLastPrimColour = 0x00000000;
DWORD			gLastEnvColour = 0x00000000;

BOOL			gLastUseZBuffer = FALSE;
bool			gNeedZBufferUdate = false;


void D3DRender::InitBlendMode()
{
	if ( gNeedZBufferUdate ||
		(gRDPOtherMode._u64 != gLastRDPOtherMode._u64) ||
		 (m_bZBuffer != gLastUseZBuffer) )
	{
		// Only update if ZBuffer is enabled
		if (m_bZBuffer)
		{
			g_pD3DDev->SetRenderState( D3DRS_ZENABLE, gRDPOtherMode.z_cmp ? D3DZB_TRUE : D3DZB_FALSE );	
			g_pD3DDev->SetRenderState( D3DRS_ZWRITEENABLE, gRDPOtherMode.z_upd ? TRUE : FALSE );	
		}
		else
		{
			g_pD3DDev->SetRenderState( D3DRS_ZENABLE, D3DZB_FALSE );	
			g_pD3DDev->SetRenderState( D3DRS_ZWRITEENABLE, FALSE );	
		}

		gNeedZBufferUdate = false;
		gLastUseZBuffer = m_bZBuffer;
	}


#ifdef AVOID_STATE_CHANGES
	if ( gRDPOtherMode._u64 == gLastRDPOtherMode._u64 &&
		 gRDPMux._u64 == gLastRDPMux._u64 &&
		 (gLastPrimColour == m_dwPrimitiveColor || !gUsingPrimColour ) &&
		 (gLastEnvColour == m_dwEnvColor || !gUsingEnvColour ) &&
		 gTextureChanged == false )
	{
		//static u32 num_ignored = 0;
		//num_ignored++;
		//if ( (num_ignored%10000) == 0)
		//	DBGConsole_Msg( 0, "Ignoring redundant state initialisation %d", num_ignored );
		return;
	}
#endif

	gLastRDPOtherMode._u64 = gRDPOtherMode._u64;
	gLastRDPMux._u64 = gRDPMux._u64;
	gTextureChanged = false;
	gLastPrimColour = m_dwPrimitiveColor;
	gLastEnvColour = m_dwEnvColor;

	switch ( gRDPOtherMode.cycle_type )
	{
		case CYCLE_COPY:			InitCycleCopy();	break;
		case CYCLE_FILL:			InitCycleFill();	break;
		case CYCLE_1CYCLE:			InitCycle1();		break;
		case CYCLE_2CYCLE:			InitCycle2();		break;
	}

	g_pD3DDev->SetRenderState( D3DRS_SHADEMODE, mSmooth ? D3DSHADE_GOURAUD : D3DSHADE_FLAT );

	//g_pD3DDev->SetRenderState( D3DRS_TEXTUREPERSPECTIVE, gRDPOtherMode.text_persp );		// Not in D3D8?
	
	switch ( gRDPOtherMode.text_filt )
	{
		case 0:			//G_TF_POINT:	// 0
			g_pD3DDev->SetTextureStageState( 0, D3DTSS_MINFILTER, D3DTEXF_POINT );
			g_pD3DDev->SetTextureStageState( 0, D3DTSS_MAGFILTER, D3DTEXF_POINT );
			g_pD3DDev->SetTextureStageState( 1, D3DTSS_MINFILTER, D3DTEXF_POINT );
			g_pD3DDev->SetTextureStageState( 1, D3DTSS_MAGFILTER, D3DTEXF_POINT );
			break;
		case 2:			//G_TF_BILERP:	// 2
			g_pD3DDev->SetTextureStageState( 0, D3DTSS_MINFILTER, D3DTEXF_LINEAR );
			g_pD3DDev->SetTextureStageState( 0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR );
			g_pD3DDev->SetTextureStageState( 1, D3DTSS_MINFILTER, D3DTEXF_LINEAR );
			g_pD3DDev->SetTextureStageState( 1, D3DTSS_MAGFILTER, D3DTEXF_LINEAR );
			break;
		case 3:			//G_TF_AVERAGE:	// 3?
			g_pD3DDev->SetTextureStageState( 0, D3DTSS_MINFILTER, D3DTEXF_LINEAR );
			g_pD3DDev->SetTextureStageState( 0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR );
			g_pD3DDev->SetTextureStageState( 1, D3DTSS_MINFILTER, D3DTEXF_LINEAR );
			g_pD3DDev->SetTextureStageState( 1, D3DTSS_MAGFILTER, D3DTEXF_LINEAR );
			break;
	}

#define BLEND_NOOP1				0x00000000
#define BLEND_NOOP2				0x00000000

#define BLEND_FOG_ASHADE1		0xc8000000
#define BLEND_FOG_APRIM1		0xc4000000

#define BLEND_PASS1				0x0c080000
#define BLEND_PASS2				0x03020000
	
#define BLEND_OPA1				0x00440000
#define BLEND_OPA2				0x00110000

#define BLEND_XLU1				0x00400000
#define BLEND_XLU2				0x00100000

#define BLEND_ADD1				0x04400000
#define BLEND_ADD2				0x01100000

#define BLEND_MEM1				0x4c400000		// Mem*0 + Mem*(1-0)?!
#define BLEND_MEM2				0x13100000		// Mem*0 + Mem*(1-0)?!


/*
#define	G_BL_CLR_IN	0
#define	G_BL_CLR_MEM	1
#define	G_BL_CLR_BL	2
#define	G_BL_CLR_FOG	3
#define	G_BL_1MA	0
#define	G_BL_A_MEM	1
#define	G_BL_A_IN	0
#define	G_BL_A_FOG	1
#define	G_BL_A_SHADE	2
#define	G_BL_1		2
#define	G_BL_0		3

#define	GBL_c1(m1a, m1b, m2a, m2b)	\
	(m1a) << 30 | (m1b) << 26 | (m2a) << 22 | (m2b) << 18
#define	GBL_c2(m1a, m1b, m2a, m2b)	\
	(m1a) << 28 | (m1b) << 24 | (m2a) << 20 | (m2b) << 16

#define	G_RM_FOG_SHADE_A		GBL_c1(G_BL_CLR_FOG, G_BL_A_SHADE, G_BL_CLR_IN, G_BL_1MA)
#define	G_RM_FOG_PRIM_A			GBL_c1(G_BL_CLR_FOG, G_BL_A_FOG, G_BL_CLR_IN, G_BL_1MA)
#define	G_RM_PASS				GBL_c1(G_BL_CLR_IN, G_BL_0, G_BL_CLR_IN, G_BL_1)
#define	RM_AA_ZB_OPA_SURF(clk)	GBL_c##clk(G_BL_CLR_IN, G_BL_A_IN, G_BL_CLR_MEM, G_BL_A_MEM)
*/
	u32 blendmode_1 = u32( gRDPOtherMode._u32[0] & 0xcccc0000 );
	u32 blendmode_2 = u32( gRDPOtherMode._u32[0] & 0x33330000 );

	if ( gRDPOtherMode.cycle_type == CYCLE_FILL )
	{
		g_pD3DDev->SetRenderState( D3DRS_SRCBLEND,   D3DBLEND_ONE);
		g_pD3DDev->SetRenderState( D3DRS_DESTBLEND,  D3DBLEND_ZERO);
		g_pD3DDev->SetRenderState( D3DRS_ALPHABLENDENABLE, FALSE );
		g_pD3DDev->SetRenderState( D3DRS_ALPHATESTENABLE, FALSE );
	}
	else if ( blendmode_1 == 0 && blendmode_2 == 0 )
	{
		g_pD3DDev->SetRenderState( D3DRS_SRCBLEND,   D3DBLEND_SRCALPHA);
		g_pD3DDev->SetRenderState( D3DRS_DESTBLEND,  D3DBLEND_INVSRCALPHA);
		g_pD3DDev->SetRenderState( D3DRS_ALPHABLENDENABLE, TRUE );
		g_pD3DDev->SetRenderState( D3DRS_ALPHATESTENABLE, TRUE );
		DL_PF( "      Blend: NOOP/NOOP" );
	}
	else if ( blendmode_1 == BLEND_FOG_ASHADE1 && blendmode_2 == BLEND_OPA2 )
	{
		// c811
		g_pD3DDev->SetRenderState( D3DRS_SRCBLEND,   D3DBLEND_ONE);
		g_pD3DDev->SetRenderState( D3DRS_DESTBLEND,  D3DBLEND_ZERO);
		g_pD3DDev->SetRenderState( D3DRS_ALPHABLENDENABLE, FALSE );
		g_pD3DDev->SetRenderState( D3DRS_ALPHATESTENABLE, FALSE );
		DL_PF( "      Blend: FOG_ASHADE/OPA" );
	}
	else if ( blendmode_1 == BLEND_FOG_ASHADE1 && blendmode_2 == BLEND_XLU2 )
	{
		// c810
		g_pD3DDev->SetRenderState( D3DRS_SRCBLEND,   D3DBLEND_SRCALPHA);
		g_pD3DDev->SetRenderState( D3DRS_DESTBLEND,  D3DBLEND_INVSRCALPHA);
		g_pD3DDev->SetRenderState( D3DRS_ALPHABLENDENABLE, TRUE );
		g_pD3DDev->SetRenderState( D3DRS_ALPHATESTENABLE, TRUE );
		DL_PF( "      Blend: FOG_ASHADE/XLU" );
	}
	else if ( blendmode_1 == BLEND_XLU1 && blendmode_2 == BLEND_XLU2 )
	{
		// 0x0050
		g_pD3DDev->SetRenderState( D3DRS_SRCBLEND,   D3DBLEND_SRCALPHA);
		g_pD3DDev->SetRenderState( D3DRS_DESTBLEND,  D3DBLEND_INVSRCALPHA);
		g_pD3DDev->SetRenderState( D3DRS_ALPHABLENDENABLE, TRUE );
		g_pD3DDev->SetRenderState( D3DRS_ALPHATESTENABLE, TRUE );
		DL_PF( "      Blend: XLU/XLU" );
	}
	else if ( blendmode_1 == BLEND_XLU1 && blendmode_2 == BLEND_NOOP2 )
	{
		// 0x0040
		g_pD3DDev->SetRenderState( D3DRS_SRCBLEND,   D3DBLEND_SRCALPHA);
		g_pD3DDev->SetRenderState( D3DRS_DESTBLEND,  D3DBLEND_INVSRCALPHA);
		g_pD3DDev->SetRenderState( D3DRS_ALPHABLENDENABLE, TRUE );
		g_pD3DDev->SetRenderState( D3DRS_ALPHATESTENABLE, TRUE );
		DL_PF( "      Blend: XLU/NOOP" );
	}
	else if ( blendmode_1 == BLEND_XLU1 && blendmode_2 == BLEND_ADD2 )
	{
		// 0x005n
		g_pD3DDev->SetRenderState( D3DRS_SRCBLEND,   D3DBLEND_ONE);
		g_pD3DDev->SetRenderState( D3DRS_DESTBLEND,  D3DBLEND_ONE);
		g_pD3DDev->SetRenderState( D3DRS_ALPHABLENDENABLE, TRUE );
		g_pD3DDev->SetRenderState( D3DRS_ALPHATESTENABLE, TRUE );
		DL_PF( "      Blend: XLU/ADD" );
	}
	else if ( blendmode_1 == BLEND_OPA1 && blendmode_2 == BLEND_OPA2 )
	{
		// 0x0055
		// Would have thoughts this was SRCALPHA/DSTALPHA, but apparently not
		g_pD3DDev->SetRenderState( D3DRS_SRCBLEND,   D3DBLEND_ONE);
		g_pD3DDev->SetRenderState( D3DRS_DESTBLEND,  D3DBLEND_ZERO);
		g_pD3DDev->SetRenderState( D3DRS_ALPHABLENDENABLE, FALSE );
		g_pD3DDev->SetRenderState( D3DRS_ALPHATESTENABLE, FALSE );
		DL_PF( "      Blend: OPA/OPA" );
	}
	else if ( blendmode_1 == BLEND_OPA1 && blendmode_2 == BLEND_NOOP2 )
	{
		// 0x0044
		g_pD3DDev->SetRenderState( D3DRS_SRCBLEND,   D3DBLEND_ONE);
		g_pD3DDev->SetRenderState( D3DRS_DESTBLEND,  D3DBLEND_ZERO);
		g_pD3DDev->SetRenderState( D3DRS_ALPHABLENDENABLE, FALSE );
		g_pD3DDev->SetRenderState( D3DRS_ALPHATESTENABLE, FALSE );
		DL_PF( "      Blend: OPA/NOOP" );
	}
	else if ( blendmode_1 == BLEND_PASS1 && blendmode_2 == BLEND_PASS2 )
	{
		// 0x0f0a
		g_pD3DDev->SetRenderState( D3DRS_SRCBLEND,   D3DBLEND_ONE);
		g_pD3DDev->SetRenderState( D3DRS_DESTBLEND,  D3DBLEND_ZERO);
		g_pD3DDev->SetRenderState( D3DRS_ALPHABLENDENABLE, FALSE );
		g_pD3DDev->SetRenderState( D3DRS_ALPHATESTENABLE, FALSE );
		DL_PF( "      Blend: PASS/PASS" );
	}
	else if ( blendmode_1 == BLEND_PASS1 && blendmode_2 == BLEND_XLU2 )
	{
		// 0x0c18
		g_pD3DDev->SetRenderState( D3DRS_SRCBLEND,   D3DBLEND_SRCALPHA);
		g_pD3DDev->SetRenderState( D3DRS_DESTBLEND,  D3DBLEND_INVSRCALPHA);
		g_pD3DDev->SetRenderState( D3DRS_ALPHABLENDENABLE, TRUE );
		g_pD3DDev->SetRenderState( D3DRS_ALPHATESTENABLE, TRUE );
		DL_PF( "      Blend: PASS/XLU" );
	}
	else if ( blendmode_1 == BLEND_PASS1 && blendmode_2 == BLEND_OPA2 )
	{
		// 0x0c19
		g_pD3DDev->SetRenderState( D3DRS_SRCBLEND,   D3DBLEND_ONE);
		g_pD3DDev->SetRenderState( D3DRS_DESTBLEND,  D3DBLEND_ZERO);
		g_pD3DDev->SetRenderState( D3DRS_ALPHABLENDENABLE, FALSE );
		g_pD3DDev->SetRenderState( D3DRS_ALPHATESTENABLE, FALSE );
		DL_PF( "      Blend: PASS/OPA" );
	}
	else if ( blendmode_1 == BLEND_MEM1 && blendmode_2 == BLEND_MEM2 )
	{
		// 0x5f50
		g_pD3DDev->SetRenderState( D3DRS_SRCBLEND,   D3DBLEND_ZERO);
		g_pD3DDev->SetRenderState( D3DRS_DESTBLEND,  D3DBLEND_ONE);
		g_pD3DDev->SetRenderState( D3DRS_ALPHABLENDENABLE, FALSE );
		g_pD3DDev->SetRenderState( D3DRS_ALPHATESTENABLE, FALSE );
		DL_PF( "      Blend: MEM/MEM" );
	}
	else
	{
		g_pD3DDev->SetRenderState( D3DRS_SRCBLEND,   D3DBLEND_SRCALPHA);
		g_pD3DDev->SetRenderState( D3DRS_DESTBLEND,  D3DBLEND_INVSRCALPHA);
		g_pD3DDev->SetRenderState( D3DRS_ALPHABLENDENABLE, TRUE );
		g_pD3DDev->SetRenderState( D3DRS_ALPHATESTENABLE, TRUE );
		DL_PF( "      Blend: SRCALPHA/INVSRCALPHA (default: 0x%04x)", gRDPOtherMode.blender );
	}
	

	//
	// I can't think why the hand in mario's menu screen is rendered with an opaque rendermode,
	// and no alpha threshold. We set the alpha reference to 1 to ensure that the transparent pixels
	// don't get rendered. I hope this doesn't fuck anything else up.
	//
	if ( gRDPOtherMode.alpha_compare == 0 )
	{
		if ( gRDPOtherMode.cvg_x_alpha )
		{
			g_pD3DDev->SetRenderState( D3DRS_ALPHAREF,        0x01 );
			g_pD3DDev->SetRenderState( D3DRS_ALPHATESTENABLE, TRUE );
		}
		else
		{
			g_pD3DDev->SetRenderState( D3DRS_ALPHATESTENABLE, FALSE );
		}
	}
	else
	{
		// G_AC_THRESHOLD || G_AC_DITHER
		g_pD3DDev->SetRenderState( D3DRS_ALPHAREF,        m_dwAlpha );
		g_pD3DDev->SetRenderState( D3DRS_ALPHATESTENABLE, TRUE );
	}

}



//*****************************************************************************
//
//*****************************************************************************
inline f32 round( f32 x )
{
	return (f32)(s32)( x + 0.5f );
}

HRESULT D3DRender::TexRect( const CVector2 & xy0, const CVector2 & xy1, const CVector2 & uv0, const CVector2 & uv1 )
{
	HRESULT hr;
	
	CVector2 screen0;
	CVector2 screen1;
	CVector2 tex_uv0( uv0 );
	CVector2 tex_uv1( uv1 );

	// Scale to Actual texture coords
	// The two cases are to handle the oversized textures hack on voodoos
	float width_scale = (m_fTileWidth[0] > m_fTexWidth[0]) ? m_fInvTileWidth[0] : m_fInvTexWidth[0];
	float height_scale = (m_fTileHeight[0] > m_fTexHeight[0]) ? m_fInvTileHeight[0] : m_fInvTexHeight[0];

	tex_uv0.x *= width_scale;
	tex_uv1.x *= width_scale;

	tex_uv0.y *= height_scale;
	tex_uv1.y *= height_scale;

	//
	// In Fill/Copy mode the coordinates are inclusive (i.e. add 1.0f to the w/h)
	//
	CVector2 edge( 0, 0 );

	switch ( gRDPOtherMode.cycle_type )
	{
		case CYCLE_COPY:			edge.x += 1.0f; edge.y += 1.0f;	break;
		case CYCLE_FILL:			edge.x += 1.0f; edge.y += 1.0f;	break;
		case CYCLE_1CYCLE:											break;
		case CYCLE_2CYCLE:											break;
	}


	//
	// We round these value here, so that when we scale up the coords to our screen
	// coords we don't get any gaps.
	//
	screen0.x = round( xy0.x * m_fScreenMultX );
	screen0.y = round( xy0.y * m_fScreenMultY );

	screen1.x = round( (xy1.x + edge.x) * m_fScreenMultX );
	screen1.y = round( (xy1.y + edge.x) * m_fScreenMultY );

#ifdef DEBUG_TEXRECT
	DL_PF( "      ScreenMult:  %f,%f", m_fScreenMultX, m_fScreenMultY );
	DL_PF( "      Screen:  %f,%f -> %f,%f", screen0.x, screen0.y, screen1.x, screen1.y );
	DL_PF( "      Texture: %f,%f -> %f,%f", tex_uv0.x, tex_uv0.y, tex_uv1.x, tex_uv1.y );
#endif

	DaedalusVtx trv[4];
	static const u16 indices[2*3] = {1,0,2, 2,0,3};

	trv[0].pos = CVector3( screen0.x, screen0.y, 1.0f );
	trv[0].rhw = 1.0f;
	trv[0].color = 0xffffffff;
	trv[0].t0 = CVector2( tex_uv0.x, tex_uv0.y );
	trv[0].t1 = trv[0].t0;

	trv[1].pos = CVector3( screen1.x, screen0.y, 1.0f );
	trv[1].rhw = 1.0f;
	trv[1].color = 0xffffffff;
	trv[1].t0 = CVector2( tex_uv1.x, tex_uv0.y );
	trv[1].t1 = trv[1].t0;

	trv[2].pos = CVector3( screen1.x, screen1.y, 1.0f );
	trv[2].rhw = 1.0f;
	trv[2].color = 0xffffffff;
	trv[2].t0 = CVector2( tex_uv1.x, tex_uv1.y );
	trv[2].t1 = trv[2].t0;

	trv[3].pos = CVector3( screen0.x, screen1.y, 1.0f );
	trv[3].rhw = 1.0f;
	trv[3].color = 0xffffffff;
	trv[3].t0 = CVector2( tex_uv0.x, tex_uv1.y );
	trv[3].t1 = trv[3].t0;

	hr = RenderTriangleList( trv, 4, indices, 6, true, true );

	return hr;
	
}

//*****************************************************************************
//
//*****************************************************************************
HRESULT D3DRender::TexRectFlip( const CVector2 & xy0, const CVector2 & xy1, const CVector2 & uv0, const CVector2 & uv1 )
{
	HRESULT hr;
	
	// TODO - In 1/2cycle mode, skip bottom/right edges!?
	CVector2 screen0;
	CVector2 screen1;
	CVector2 tex_uv0( uv0 );
	CVector2 tex_uv1( uv1 );

	// Scale to Actual texture coords
	// The two cases are to handle the oversized textures hack on voodoos
	float width_scale = (m_fTileWidth[0] > m_fTexWidth[0]) ? m_fInvTileWidth[0] : m_fInvTexWidth[0];
	float height_scale = (m_fTileHeight[0] > m_fTexHeight[0]) ? m_fInvTileHeight[0] : m_fInvTexHeight[0];

	tex_uv0.x *= width_scale;
	tex_uv1.x *= width_scale;

	tex_uv0.y *= height_scale;
	tex_uv1.y *= height_scale;

	//
	// In Fill/Copy mode the coordinates are inclusive (i.e. add 1.0f to the w/h)
	//
	CVector2 edge( 0, 0 );

	switch ( gRDPOtherMode.cycle_type )
	{
		case CYCLE_COPY:			edge.x += 1.0f; edge.y += 1.0f;	break;
		case CYCLE_FILL:			edge.x += 1.0f; edge.y += 1.0f;	break;
		case CYCLE_1CYCLE:											break;
		case CYCLE_2CYCLE:											break;
	}

	//
	// We round these value here, so that when we scale up the coords to our screen
	// coords we don't get any gaps.
	//
	screen0.x = round( xy0.x * m_fScreenMultX );
	screen0.y = round( xy0.y * m_fScreenMultY );

	screen1.x = round( (xy1.x + edge.x) * m_fScreenMultX );
	screen1.y = round( (xy1.y + edge.x) * m_fScreenMultY );

#ifdef DEBUG_TEXRECT
	DL_PF( "      ScreenMult:  %f,%f", m_fScreenMultX, m_fScreenMultY );
	DL_PF( "      Screen:  %f,%f -> %f,%f", screen0.x, screen0.y, screen1.x, screen1.y );
	DL_PF( "      Texture: %f,%f -> %f,%f", tex_uv0.x, tex_uv0.y, tex_uv1.x, tex_uv1.y );
#endif

	DaedalusVtx trv[4];
	static const u16 indices[2*3] = {1,0,2, 2,0,3};

	// Same as TexRect, but with texcoords 0,2 swapped
	trv[0].pos = CVector3( screen0.x, screen0.y, 1.0f );
	trv[0].rhw = 1.0f;
	trv[0].color = 0xffffffff;
	trv[0].t0 = CVector2( tex_uv0.x, tex_uv0.y );
	trv[0].t1 = trv[0].t0;
	
	trv[1].pos = CVector3( screen1.x, screen0.y, 1.0f );
	trv[1].rhw = 1.0f;
	trv[1].color = 0xffffffff;
	trv[1].t0 = CVector2( tex_uv0.x, tex_uv1.y );		// In TexRect this is tex_uv1.x, tex_uv0.y
	trv[1].t1 = trv[1].t0;

	trv[2].pos = CVector3( screen1.x, screen1.y, 1.0f );
	trv[2].rhw = 1.0f;
	trv[2].color = 0xffffffff;
	trv[2].t0 = CVector2( tex_uv1.x, tex_uv1.y );
	trv[2].t1 = trv[2].t0;

	trv[3].pos = CVector3( screen0.x, screen1.y, 1.0f );
	trv[3].rhw = 1.0f;
	trv[3].color = 0xffffffff;
	trv[3].t0 = CVector2( tex_uv1.x, tex_uv0.y );		//tex_uv0.x	tex_uv1.y	
	trv[3].t1 = trv[3].t0;

	hr = RenderTriangleList( trv, 4, indices, 6, true, true );

	return hr;
}

//*****************************************************************************
//
//*****************************************************************************
HRESULT D3DRender::FillRect( const CVector2 & xy0, const CVector2 & xy1, u32 color )
{
	HRESULT hr;

	CVector2 screen0;
	CVector2 screen1;

	if ( (gRDPOtherMode._u32[0] & 0xffff0000) == 0x5f500000 )
	{
		// this blend mode is mem*0 + mem*1, so we don't need to render it... Very odd!
		return S_OK;
	}

	// This if for C&C - It might break other stuff (I'm not sure if we should allow alpha or not..)
//	color |= 0xff000000;

	//
	// We round these value here, so that when we scale up the coords to our screen
	// coords we don't get any gaps.
	//
	screen0.x = round( xy0.x * m_fScreenMultX );
	screen0.y = round( xy0.y * m_fScreenMultY );

	screen1.x = round( xy1.x * m_fScreenMultX );
	screen1.y = round( xy1.y * m_fScreenMultY );

	DL_PF( "      ScreenMult:  %f,%f", m_fScreenMultX, m_fScreenMultY );
	DL_PF( "      Screen:  %f,%f -> %f,%f", screen0.x, screen0.y, screen1.x, screen1.y );

	DaedalusVtx frv[4];
	static const u16 indices[2*3] = {1,0,2, 2,0,3};

	frv[0].pos = CVector3(screen0.x, screen0.y, 1 );
	frv[0].rhw = 1.0f;
	frv[0].color = color;
	frv[0].t0 = CVector2( 0, 0 );
	frv[0].t1 = frv[0].t0;

	frv[1].pos = CVector3(screen1.x, screen0.y, 1 );
	frv[1].rhw = 1.0f;
	frv[1].color = color;
	frv[1].t0 = CVector2( 1, 0 );
	frv[1].t1 = frv[1].t0;

	frv[2].pos = CVector3(screen1.x, screen1.y, 1 );
	frv[2].rhw = 1.0f;
	frv[2].color = color;
	frv[2].t0 = CVector2( 1, 1 );
	frv[2].t1 = frv[2].t0;

	frv[3].pos = CVector3(screen0.x, screen1.y, 1 );
	frv[3].rhw = 1.0f;
	frv[3].color = color;
	frv[3].t0 = CVector2( 0, 1 );
	frv[3].t1 = frv[3].t0;

	hr = RenderTriangleList( frv, 4, indices, 6, true );
	
	return hr;
}


//*****************************************************************************
//
//*****************************************************************************
// Returns TRUE if it thinks the triangle is visible
// Returns FALSE if it is clipped
bool D3DRender::TestTri(u32 v0, u32 v1, u32 v2) const
{
	//
	// Render all the triangles while exporting
	//
#ifdef VRML_EXPORT_SUPPORT
	return true;
#endif
	// Check vertices are valid!
	if ( v0 >= MAX_VERTS || v1 >= MAX_VERTS || v2 >= MAX_VERTS )
		return false;

	// Here we AND all the flags. If any of the bits is set for all
	// 3 vertices, it means that all three x, y or z lie outside of
	// the current viewing volume.
	// Currently disabled - still seems a bit dodgy
#if CLIPPING
	if (m_dwVecFlags[v0] & m_dwVecFlags[v1] & m_dwVecFlags[v2])
	{
		//DL_PF("0x%04x 0x%04x 0x%04x", m_dwVecFlags[dwV0], m_dwVecFlags[dwV1], m_dwVecFlags[dwV2]);
		return false;
	}
	else
#endif
	if ((m_bCullFront || m_bCullBack) && !IsZModeDecal())
	{
		const D3DXVECTOR3 & vert0 = m_VtxTransformed[v0].pos;
		const D3DXVECTOR3 & vert1 = m_VtxTransformed[v1].pos;
		const D3DXVECTOR3 & vert2 = m_VtxTransformed[v2].pos;

		// Only try to clip if the tri is onscreen. For some reason, this
		// method doesnt' work well when the z value is outside of screenspace
		if (vert0.z < 1 && vert1.z < 1 && vert2.z < 1)
		{
			float V1 = vert2.x - vert0.x;
			float V2 = vert2.y - vert0.y;
			
			float W1 = vert2.x - vert1.x;
			float W2 = vert2.y - vert1.y;

			float fDirection = -( (V1 * W2) - (V2 * W1) );

			if (fDirection < 0 && m_bCullBack)
				return false;
			else if (fDirection > 0 && m_bCullFront)
				return false;
		}
	}
	return true;
}


//*****************************************************************************
//
//*****************************************************************************
bool D3DRender::TestVerts( u32 v0, u32 vn ) const
{
	u32 flags = m_dwVecFlags[v0];

	for ( u32 i = v0+1; i <= vn && i < MAX_VERTS; i++ )
	{
		flags &= m_dwVecFlags[i];
	}

	return flags == 0;
}


//*****************************************************************************
//
//*****************************************************************************
DWORD D3DRender::LightVert( const D3DXVECTOR3 & norm )
{
	float r,g,b;
	float fCosT;
	s32 l;

	// Ip = Light intensity
	// Ia = Ambient light intensity
	// Ka = Material ambient reflective property
	// Kd = Material diffuse reflective property
	// L = normalised light direction
	// N = vertex normal

	// I = IaKa + IpKd (N . L)

	// Do ambient
	r = m_AmbientCol.r;
	g = m_AmbientCol.g;
	b = m_AmbientCol.b;
	//a = m_fAmbientA;

	for ( l = 0; l < m_dwNumLights; l++ )
	{
		fCosT = D3DXVec3Dot(&norm, &m_Lights[l].dir);
			
		if (fCosT > 0)
		{
			r += m_Lights[l].col.r * fCosT;
			g += m_Lights[l].col.g * fCosT;
			b += m_Lights[l].col.b * fCosT;
		//	a += m_Lights[l].col.a * fCosT;
		}
	}

	if (r > 255.0f) r = 255.0f;
	if (g > 255.0f) g = 255.0f;
	if (b > 255.0f) b = 255.0f;
	//if (a > 255.0f) a = 255.0f;

	return D3DCOLOR_RGBA((LONG)r, (LONG)g, (LONG)b, (LONG)255);

}


//*****************************************************************************
//
//*****************************************************************************
HRESULT D3DRender::InitialiseBuffers( DWORD * pvtxbase, DWORD * pidxbase )
{
	HRESULT hr;

	DaedalusVtx * vertices;
	WORD * indices;

	if ( m_lpVtxBuffer == NULL || m_lpIdxBuffer == NULL )
	{
		RestoreDeviceObjects();

		// If we couldn't create, return S_FALSE for now (assume buffers will be created later
		if (m_lpVtxBuffer == NULL || m_lpIdxBuffer == NULL )
			return S_FALSE;
	}

	// From the DX8 SDK PointSprite sample. Lock the VB in chunks, 
	// using NOOVERWRITE where possible so that the device isn't forced to
	// stall until the vertices are rendered. Then the base pointer wraps around
	// it is reset to 0 and the buffer is discarded
	if ( m_dwVtxBufBase + m_dwNumVertices >= m_dwVtxBufSize )
		m_dwVtxBufBase = 0;
	if ( m_dwIdxBufBase + m_dwNumIndices >= m_dwIdxBufSize )
		m_dwIdxBufBase = 0;

	//DBGConsole_Msg(0, "Base: %d, Len: %d, Num: %d", m_dwVtxBufBase, m_dwVtxBufSize, m_dwNumVertices );

	//
	// Copy vertices
	//
	hr = m_lpVtxBuffer->Lock( m_dwVtxBufBase  * sizeof( DaedalusVtx ),		// Start
							  m_dwNumVertices * sizeof( DaedalusVtx ),		// Length
							  (BYTE**)&vertices,
							  m_dwVtxBufBase ? D3DLOCK_NOOVERWRITE : D3DLOCK_DISCARD );

	if (FAILED(hr))
	{
		m_dwNumVertices = 0;
		m_dwNumIndices = 0;
		return hr;
	}

	memcpy( vertices, m_swVertexBuffer, m_dwNumVertices * sizeof( DaedalusVtx ) );

	m_lpVtxBuffer->Unlock();

	
	//
	// Copy indices
	//
	hr = m_lpIdxBuffer->Lock( m_dwIdxBufBase * sizeof( WORD ),
							  m_dwNumIndices * sizeof( WORD ),
							  (BYTE**)&indices, 
							  m_dwIdxBufBase ? D3DLOCK_NOOVERWRITE : D3DLOCK_DISCARD );
	if (FAILED(hr))
	{
		m_lpVtxBuffer->Unlock();
		m_dwNumVertices = 0;
		m_dwNumIndices = 0;
		return hr;
	}

	memcpy( indices, m_swIndexBuffer, m_dwNumIndices * sizeof( WORD ) );

	m_lpIdxBuffer->Unlock();

	*pvtxbase = m_dwVtxBufBase;
	*pidxbase = m_dwIdxBufBase;

	m_dwVtxBufBase += m_dwNumVertices;
	m_dwIdxBufBase += m_dwNumIndices;

	return S_OK;

}


//*****************************************************************************
//
//*****************************************************************************
#ifdef VRML_EXPORT_SUPPORT
void D3DRender::DumpVerticesToSceneFile()
{
	SceneObject so;
	m_arrObjects.push_back( so );

	u32 index = m_arrObjects.size() - 1;

#ifdef LIBPNG_SUPPORT
	CHAR szFileName[ MAX_PATH + 1 ] = "";
	if ( m_lpsTexture[0] )
	{
		CHAR szFile[ MAX_PATH + 1 ];
		sprintf( szFile, "c:\\temp\\Textures\\0x%08x.png", m_lpsTexture[0] );
		LPDIRECT3DSURFACE8 pSurf = NULL;

		m_lpsTexture[0]->GetSurfaceLevel( 0, &pSurf );

		if ( pSurf )
		{
			PngSaveSurface( szFile, pSurf );

			pSurf->Release();

			// Relative url
			sprintf( szFileName, "0x%08x.png", m_lpsTexture[0] );
		}
	}
	m_arrObjects[ index ].SetTextureURL( szFileName );
#endif // LIBPNG_SUPPORT
	for ( u32 i = 0; i < m_dwNumIndices; i ++ )
	{
		DaedalusVtx v = m_SceneVerts[ m_swIndexBuffer[ i ] ];

		// Override the position with the position before the projection
		// matrix was applied.
		v.pos = m_vecPreProjected[ m_swIndexBuffer[ i ] ];
		m_arrObjects[ index ].AddVertex( v );
	}

}
#endif

//*****************************************************************************
//
//*****************************************************************************
HRESULT D3DRender::FlushTris()
{
	HRESULT hr;

	if ( m_dwNumIndices == 0 )
		return S_OK;

	//
	// Set up the vertices. We could optimise this by re-indexing the vertices
	// so that we only send the used vertices to the card. Here we check 
	// for the greatest index used, and use this for our vertex number
	//
	u32 min_vert_index = MAX_VERTS;
	u32 max_vert_index = 0;
	for ( u32 i = 0; i < m_dwNumIndices; i++ )
	{
		if ( m_swIndexBuffer[ i ] > max_vert_index )
			max_vert_index = m_swIndexBuffer[ i ];
		if ( m_swIndexBuffer[ i ] < min_vert_index )
			min_vert_index = m_swIndexBuffer[ i ];
	}

	if ( min_vert_index > 0 )
	{
		//DBGConsole_Msg( 0, "Min index is %d (max is %d)", min_vert_index, max_vert_index );
		for ( i = 0; i < m_dwNumIndices; i++ )
		{
			m_swIndexBuffer[ i ] -= (u16)min_vert_index;
		}

		max_vert_index -= min_vert_index;
	}

	m_dwNumVertices = max_vert_index + 1;


	if ( m_dwNumVertices == 0 )
		return S_OK;

	memcpy( m_swVertexBuffer, &m_VtxTransformed[ min_vert_index ], m_dwNumVertices * sizeof( DaedalusVtx ) );

	//
	// Process the software vertex buffer to apply a couple of 
	// necessary changed to the texture coords (this is required
	// because some ucodes set the texture after setting the vertices)
	//
	if (m_bTextureEnable)
	{
		// Only scale if not generated automatically
		if ( !( m_bTextureGen && m_bLightingEnable ) )
		{
			for ( u32 v = 0; v < m_dwNumVertices; v++ )
			{
				m_swVertexBuffer[v].t0.x = (m_swVertexBuffer[v].t0.x - m_fTexLeft[0]) * m_fInvTexWidth[0];
				m_swVertexBuffer[v].t0.y = (m_swVertexBuffer[v].t0.y - m_fTexTop[0]) * m_fInvTexHeight[0];

				m_swVertexBuffer[v].t1.x = (m_swVertexBuffer[v].t1.x - m_fTexLeft[1]) * m_fInvTexWidth[1];
				m_swVertexBuffer[v].t1.y = (m_swVertexBuffer[v].t1.y - m_fTexTop[1]) * m_fInvTexHeight[1];
			}
		}

		// Bias points in decal mode
		if (IsZModeDecal())
		{
			for ( u32 v = 0; v < m_dwNumVertices; v++ )
			{
				m_swVertexBuffer[v].pos.z -= 0.0003f;
			}
		}
	}

	//
	// If smooth shading is turned off, duplicate the first diffuse color for all three verts
	//
	if ( !mSmoothShade )
	{

	}


#ifdef VRML_EXPORT_SUPPORT
	if ( m_DumpingScene )
	{	
		DumpVerticesToSceneFile();
	}
#endif // VRML_EXPORT_SUPPORT


	START_PROFILE( CProfiler::PROFILE_GFX_COMBINER );
	InitBlendMode();
	STOP_PROFILE( CProfiler::PROFILE_GFX_COMBINER );

	DWORD vtxbase, idxbase;

	hr = InitialiseBuffers( &vtxbase, &idxbase );
	if ( SUCCEEDED(hr) )
	{
		g_pD3DDev->SetVertexShader( FVF_DaedalusVtx );
		g_pD3DDev->SetIndices( m_lpIdxBuffer, vtxbase );
		g_pD3DDev->SetStreamSource( 0, m_lpVtxBuffer, sizeof( DaedalusVtx ) );

#ifndef USE_PIXEL_SHADERS
		if ( gRDPMux._u64 == 0x00127e2433fdf8fc )
		{
			hr = RenderWaterHack( vtxbase, idxbase );
		}
		else if ( gRDPMux._u64 == 0x00127e44fffff238 )
		{
			hr = RenderNSHack( vtxbase, idxbase );
		}
		else
#endif
		{
			hr = g_pD3DDev->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,
												 vtxbase,
												 m_dwNumVertices,
												 idxbase,
												 m_dwNumIndices / 3 );
		}
	}

	// Reset index
	m_dwNumVertices = 0;
	m_dwNumIndices = 0;

	return hr;

}

#ifndef USE_PIXEL_SHADERS
//*****************************************************************************
//
//*****************************************************************************
HRESULT D3DRender::RenderWaterHack( DWORD vtxbase, DWORD idxbase )
{
	HRESULT hr;

	// aRGB0: (Texel0       - Prim        ) * Shade        + Primitive
	// aA0  : (0            - 0           ) * 0            + Shade
	// aRGB1: (Texel0       - Prim        ) * Shade        + Primitive
	// aA1  : (0            - 0           ) * 0            + Shade


	//(T0 - Prim) * Shade + Prim,   Shade

	// TEXEL0,PRIM,SHADE,PRIM  0,0,0,SHADE -- Used in Wetrix, incorrect at mo
	// (T0 - PRIM) * DIFFUSE + PRIM

	g_pD3DDev->SetRenderState( D3DRS_ALPHABLENDENABLE,   TRUE);
	g_pD3DDev->SetRenderState( D3DRS_ALPHATESTENABLE,   TRUE);

	g_pD3DDev->SetTextureStageState( 1, D3DTSS_COLOROP,   D3DTOP_DISABLE );
	g_pD3DDev->SetTextureStageState( 1, D3DTSS_ALPHAOP,   D3DTOP_DISABLE );

	// T0 - PRIM
	// SHADE
	g_pD3DDev->SetTexture( 0, m_lpsTexture[0] );
	g_pD3DDev->SetTexture( 1, m_lpsTexture[1] );
	SetTextureFactor(m_dwPrimitiveColor);
	
	g_pD3DDev->SetTextureStageState( 0, D3DTSS_COLOROP, D3DTOP_MODULATE );	
	g_pD3DDev->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TEXTURE );
	g_pD3DDev->SetTextureStageState( 0, D3DTSS_COLORARG2, D3DTA_DIFFUSE );
	
	g_pD3DDev->SetTextureStageState( 0, D3DTSS_ALPHAOP, D3DTOP_SELECTARG2 );
	g_pD3DDev->SetTextureStageState( 0, D3DTSS_ALPHAARG2, D3DTA_TFACTOR );		// DIFFUSE

	hr = g_pD3DDev->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,
										 vtxbase,
										 m_dwNumVertices,
										 idxbase,
										 m_dwNumIndices / 3 );		// Num prims


	// CURRENT * DIFFUSE
	g_pD3DDev->SetRenderState( D3DRS_SRCBLEND,   D3DBLEND_ONE);
	g_pD3DDev->SetRenderState( D3DRS_DESTBLEND,  D3DBLEND_ONE);

	g_pD3DDev->SetTexture( 0, NULL );

	g_pD3DDev->SetTextureStageState( 0, D3DTSS_COLOROP, D3DTOP_MODULATE );	
	g_pD3DDev->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_COMPLEMENT|D3DTA_DIFFUSE );		// 1 - DIFFUSE
	g_pD3DDev->SetTextureStageState( 0, D3DTSS_COLORARG2, D3DTA_TFACTOR );

	
	//g_pD3DDev->SetTextureStageState( 0, D3DTSS_ALPHAOP, D3DTOP_DISABLE );
	g_pD3DDev->SetTextureStageState( 0, D3DTSS_ALPHAOP, D3DTOP_SELECTARG2 );
	g_pD3DDev->SetTextureStageState( 0, D3DTSS_ALPHAARG2, D3DTA_TFACTOR );		// DIFFUSE

	hr = g_pD3DDev->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,
										 vtxbase,
										 m_dwNumVertices,
										 idxbase,
										 m_dwNumIndices / 3 );		// Num prims

	// Restore blend mode
	g_pD3DDev->SetRenderState( D3DRS_SRCBLEND,   D3DBLEND_SRCALPHA);
	g_pD3DDev->SetRenderState( D3DRS_DESTBLEND,  D3DBLEND_INVSRCALPHA);


	return hr;
}
#endif //USE_PIXEL_SHADERS

#ifndef USE_PIXEL_SHADERS
//*****************************************************************************
//
//*****************************************************************************
HRESULT D3DRender::RenderNSHack( DWORD vtxbase, DWORD idxbase )
{
	HRESULT hr;

	// Spiderman/TH
	//RGB0: (Texel0       - 0           ) * Shade        + 0           
	//A0  : (0            - 0           ) * 0            + Texel0      
	//RGB1: (Texel1       - 0           ) * Shade        + Combined    
	//A1  : (0            - 0           ) * 0            + Combined  

	g_pD3DDev->SetRenderState( D3DRS_ALPHABLENDENABLE,   TRUE);
	g_pD3DDev->SetRenderState( D3DRS_ALPHATESTENABLE,   TRUE);

	g_pD3DDev->SetRenderState( D3DRS_SRCBLEND,   D3DBLEND_ONE);
	g_pD3DDev->SetRenderState( D3DRS_DESTBLEND,  D3DBLEND_ZERO);

	g_pD3DDev->SetTextureStageState( 1, D3DTSS_COLOROP,   D3DTOP_DISABLE );
	g_pD3DDev->SetTextureStageState( 1, D3DTSS_ALPHAOP,   D3DTOP_DISABLE );

	// T0 * SHADE
	// T0
	g_pD3DDev->SetTexture( 0, m_lpsTexture[0] );
	g_pD3DDev->SetTexture( 1, m_lpsTexture[1] );
	
	g_pD3DDev->SetTextureStageState( 0, D3DTSS_COLOROP, D3DTOP_MODULATE );	
	g_pD3DDev->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TEXTURE );
	g_pD3DDev->SetTextureStageState( 0, D3DTSS_COLORARG2, D3DTA_DIFFUSE );
	
	g_pD3DDev->SetTextureStageState( 0, D3DTSS_ALPHAOP, D3DTOP_SELECTARG1 );
	g_pD3DDev->SetTextureStageState( 0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE );

	hr = g_pD3DDev->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,
										 vtxbase,
										 m_dwNumVertices,
										 idxbase,
										 m_dwNumIndices / 3 );		// Num prims



	// T1 * SHADE
	// T0
	g_pD3DDev->SetRenderState( D3DRS_SRCBLEND,   D3DBLEND_ONE);
	g_pD3DDev->SetRenderState( D3DRS_DESTBLEND,  D3DBLEND_ONE);

	//g_pD3DDev->SetTexture( 0, m_lpsTexture[0] );
	//g_pD3DDev->SetTexture( 1, m_lpsTexture[1] );

	g_pD3DDev->SetTextureStageState( 0, D3DTSS_COLOROP, D3DTOP_MODULATE );	
	g_pD3DDev->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TEXTURE );
	g_pD3DDev->SetTextureStageState( 0, D3DTSS_COLORARG2, D3DTA_DIFFUSE );

	
	g_pD3DDev->SetTextureStageState( 0, D3DTSS_ALPHAOP, D3DTOP_SELECTARG1 );
	g_pD3DDev->SetTextureStageState( 0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE );

	hr = g_pD3DDev->DrawIndexedPrimitive(D3DPT_TRIANGLELIST,
										 vtxbase,
										 m_dwNumVertices,
										 idxbase,
										 m_dwNumIndices / 3 );		// Num prims


	g_pD3DDev->SetRenderState( D3DRS_SRCBLEND,   D3DBLEND_SRCALPHA);
	g_pD3DDev->SetRenderState( D3DRS_DESTBLEND,  D3DBLEND_INVSRCALPHA);

	return hr;
}
#endif //USE_PIXEL_SHADERS



//*****************************************************************************
// Used for TexRect, TexRectFlip, FillRect
//*****************************************************************************
HRESULT	D3DRender::RenderTriangleList( const DaedalusVtx * p_verts, u32 num_verts, const u16 * p_indices, u32 num_indices, bool disable_zbuffer, bool force_point_filter )
{
	HRESULT hr;
	
	// Save ZBuffer state
#if 0
	DWORD dwMinFilter;
	DWORD dwMaxFilter;
#endif

	if ( disable_zbuffer )
	{
		gNeedZBufferUdate = true;
	}

	if ( force_point_filter )
	{
#if 0
		// Mariokart hack to make bg on main screen work well Can do this or just clamp values
		g_pD3DDev->GetTextureStageState( 0, D3DTSS_MINFILTER, &dwMinFilter );
		g_pD3DDev->GetTextureStageState( 0, D3DTSS_MAGFILTER, &dwMaxFilter );
#endif
	}
	
	InitBlendMode();

	if ( disable_zbuffer )
	{
		g_pD3DDev->SetRenderState( D3DRS_ZENABLE, D3DZB_FALSE );	
		g_pD3DDev->SetRenderState( D3DRS_ZWRITEENABLE, FALSE );
	}

	if ( force_point_filter )
	{
		// Mariokart hack to make bg on main screen work well Can do this or just clamp values
		g_pD3DDev->SetTextureStageState( 0, D3DTSS_MINFILTER, D3DTEXF_POINT );
		g_pD3DDev->SetTextureStageState( 0, D3DTSS_MAGFILTER, D3DTEXF_POINT );
	}


	g_pD3DDev->SetVertexShader( FVF_DaedalusVtx );
	hr = g_pD3DDev->DrawIndexedPrimitiveUP(D3DPT_TRIANGLELIST,
										   0, num_verts,
										   num_indices / 3,		// 2 prims, 6 indices
										   p_indices, 
										   D3DFMT_INDEX16, 
										   p_verts, 
										   sizeof(p_verts[0]) );
	if ( force_point_filter )
	{
#if 0
		g_pD3DDev->SetTextureStageState( 0, D3DTSS_MINFILTER, dwMinFilter );
		g_pD3DDev->SetTextureStageState( 0, D3DTSS_MAGFILTER, dwMaxFilter );
#endif
	}

	return hr;
}



//*****************************************************************************
//
//*****************************************************************************
void D3DRender::SetNewVertexInfo(DWORD dwAddress, DWORD dwV0, DWORD dwNum)
{
	FiddledVtx * pVtxBase = (FiddledVtx*)(g_pu8RamBase + dwAddress);

	D3DXVECTOR4 vec4;

	D3DXMATRIX matWorldProject = m_mModelView[m_dwModelViewTop] * m_mProjection[m_dwProjectionTop];

	DWORD i;
	for (i = dwV0; i < dwV0 + dwNum; i++)
	{
		D3DXVECTOR3 vecProjected;
		D3DXVECTOR3 vecTransformedNormal;		// Used only when m_bLightingEnable
		D3DXVECTOR3 w, wp;
		u32 clip_flags;

		FiddledVtx & vert = pVtxBase[i - dwV0];

		w.x = (float)vert.x;
		w.y = (float)vert.y;
		w.z = (float)vert.z;

#ifdef VRML_EXPORT_SUPPORT
		D3DXVec3TransformCoord( &m_vecPreProjected[i], &w, &m_mModelView[m_dwModelViewTop] );
#endif

		D3DXVec3Transform(&vec4, &w, &matWorldProject);	// Convert to w=1

		static bool use_abs = false;

		m_VtxTransformed[i].rhw = 1.0f / ( vec4.w );

		vecProjected.x = vec4.x * m_VtxTransformed[i].rhw;
		vecProjected.y = vec4.y * m_VtxTransformed[i].rhw;
		vecProjected.z = vec4.z * m_VtxTransformed[i].rhw;

		// Convert to screen coords here
		m_VtxTransformed[i].pos.x = ( vecProjected.x * mVpScale.x ) + mVpTrans.x;
		m_VtxTransformed[i].pos.y = ( vecProjected.y * mVpScale.y ) + mVpTrans.y;
		m_VtxTransformed[i].pos.z = ( vecProjected.z * mVpScale.z ) + mVpTrans.z;

		clip_flags = 0;
#if CLIPPING
		D3DXVECTOR3 vecW1( vec4.x, vec4.y, vec4.z );

		vecW1 /= ( vec4.w );

		if (vecW1.x > 1.0f) clip_flags |= X_POS;
		else if (vecW1.x < -1.0f) clip_flags |= X_NEG;

		if (vecW1.y > 1.0f) clip_flags |= Y_POS;
		else if (vecW1.y < -1.0f) clip_flags |= Y_NEG;

		if (vecW1.z > 1.0f) clip_flags |= Z_POS;
//		else if (vecW1.z < -1.0f) clip_flags |= Z_NEG;		// NoN
#endif

		m_dwVecFlags[i] = clip_flags;

		if (m_bLightingEnable)
		{
			D3DXVECTOR3 mn;		// modelnorm

			mn.x = (float)vert.norma.nx;
			mn.y = (float)vert.norma.ny;
			mn.z = (float)vert.norma.nz;

			D3DXVec3TransformNormal(&vecTransformedNormal, &mn, &m_mModelView[m_dwModelViewTop]);

			// Do lighting 
			D3DXVec3Normalize(&vecTransformedNormal, &vecTransformedNormal);

			m_VtxTransformed[i].color = LightVert(vecTransformedNormal);
		}
		else
		{
			m_VtxTransformed[i].color = D3DCOLOR_RGBA( vert.rgba.r, vert.rgba.g, vert.rgba.b, vert.rgba.a );
		}

		// I decided to take this out for v0.06b release - it's not quite
		// stable enough yet
		if (m_bFogEnable)
		{
			float eyespace_z = vec4.z / fabsf( vec4.w );
			float fFog = (eyespace_z * m_fFogMult) + m_fFogOffset;

			s32 fog_coeff = (s32)(fFog * 255.0f);

			s32 alpha = max( min( fog_coeff, 255 ), 0 );

			// Set the alpha
			m_VtxTransformed[i].color &= 0x00ffffff;
			m_VtxTransformed[i].color |= alpha << 24;
		}


		if (m_bTextureEnable)
		{
			// Update texture coords n.b. need to divide tu/tv by bogus scale on addition to buffer

			// If the vert is already lit, then there is no normal (and hence we
			// can't generate tex coord)
			if (m_bTextureGen && m_bLightingEnable)
			{
				D3DXMATRIX & matWV = m_mModelView[m_dwModelViewTop];
				D3DXVECTOR3 & norm = vecTransformedNormal;

				// Assign the spheremap's texture coordinates
				m_VtxTransformed[i].t0.x = (0.5f * ( 1.0f + ( norm.x*matWV._11 +
						                 norm.y*matWV._21 +
										 norm.z*matWV._31 ) ))/* * m_fTexScaleX*/;

				m_VtxTransformed[i].t0.y = (0.5f * ( 1.0f - ( norm.x*matWV._12 +
						                 norm.y*matWV._22 +
										 norm.z*matWV._32 ) ))/* * m_fTexScaleY*/;
			}
			else
			{
				m_VtxTransformed[i].t0.x = (float)vert.tu * m_fTexScaleX;
				m_VtxTransformed[i].t0.y = (float)vert.tv * m_fTexScaleY;
			}

			m_VtxTransformed[i].t1 = m_VtxTransformed[i].t0;
		}
	}
	
}


//*****************************************************************************
// Assumes dwAddress has already been checked!	
// Don't inline - it's too big with the transform macros
// DKR seems to use longer vert info
//*****************************************************************************
void D3DRender::SetNewVertexInfoDKR(DWORD dwAddress, DWORD dwV0, DWORD dwNum)
{
	LONG nFogR, nFogG, nFogB, nFogA;
	LONG nR, nG, nB, nA;

	if (m_bFogEnable)
	{
		nFogR = RGBA_GETRED(m_dwFogColor);
		nFogG = RGBA_GETGREEN(m_dwFogColor);
		nFogB = RGBA_GETBLUE(m_dwFogColor);
		nFogA = RGBA_GETALPHA(m_dwFogColor);
	}


	s16 * pVtxBase = (s16*)(g_pu8RamBase + dwAddress);

	D3DXVECTOR4 vec4;

	D3DXMATRIX matWorldProject = m_mModelView[m_dwModelViewTop] * m_mProjection[m_dwProjectionTop];

	DWORD i;
	LONG nOff;

	nOff = 0;
	for (i = dwV0; i < dwV0 + dwNum; i++)
	{
		D3DXVECTOR3 w, wp;
		DWORD dwFlags;


		w.x = (float)pVtxBase[(nOff + 0) ^ 1];
		w.y = (float)pVtxBase[(nOff + 1) ^ 1];
		w.z = (float)pVtxBase[(nOff + 2) ^ 1];

		D3DXVec3Transform(&vec4, &w, &matWorldProject);	// Convert to w=1
		m_VtxTransformed[i].rhw = 1.0f / vec4.w;

		D3DXVECTOR3 vecProjected;

		vecProjected.x = vec4.x * m_VtxTransformed[i].rhw;
		vecProjected.y = vec4.y * m_VtxTransformed[i].rhw;
		vecProjected.z = vec4.z * m_VtxTransformed[i].rhw;

		// Convert to screen coords here
		m_VtxTransformed[i].pos.x = ( vecProjected.x * mVpScale.x ) + mVpTrans.x;
		m_VtxTransformed[i].pos.y = ( vecProjected.y * mVpScale.y ) + mVpTrans.y;
		m_VtxTransformed[i].pos.z = ( vecProjected.z * mVpScale.z ) + mVpTrans.z;

		dwFlags = 0;

		// Clip
		m_dwVecFlags[i] = dwFlags;

		s16 wA = pVtxBase[(nOff + 3) ^ 1];
		s16 wB = pVtxBase[(nOff + 4) ^ 1];

		s8 r = (s8)(wA >> 8);
		s8 g = (s8)(wA);
		s8 b = (s8)(wB >> 8);
		s8 a = (s8)(wB);

		D3DXVECTOR3 vecTransformedNormal;

		if (m_bLightingEnable)
		{
			D3DXVECTOR3 mn;		// modelnorm
			DWORD dwCol;
			D3DXMATRIX matWorld = m_mModelView[m_dwModelViewTop];

			mn.x = (float)r; //norma.nx;
			mn.y = (float)g; //norma.ny;
			mn.z = (float)b; //norma.nz;

			D3DXVec3TransformNormal(&vecTransformedNormal, &mn, &matWorld);

			// Do lighting 
			D3DXVec3Normalize(&vecTransformedNormal, &vecTransformedNormal);

			dwCol = LightVert(vecTransformedNormal);

			nR = RGBA_GETRED(dwCol);
			nG = RGBA_GETGREEN(dwCol);
			nB = RGBA_GETBLUE(dwCol);
			nA = a;//RGBA_GETALPHA(dwCol);
		}
		else
		{
			nR = r;
			nG = g;
			nB = b;
			nA = a;
		}

		// Assign true vert colour after lighting/fogging
		m_VtxTransformed[i].color = D3DCOLOR_RGBA(nR, nG, nB, nA);

		/*if (m_bTextureEnable)
		{
			// Update texture coords n.b. need to divide tu/tv by bogus scale on addition to buffer
			D3DXVECTOR2 & t = m_vecTexture[i];

			// If the vert is already lit, then there is no normal (and hence we
			// can't generate tex coord)
			if (m_bTextureGen && m_bLightingEnable)
			{
				D3DXMATRIX & matWV = m_mModelView[m_dwModelViewTop];
				D3DXVECTOR3 & norm = vecTransformedNormal;

				// Assign the spheremap's texture coordinates
				t.x = (0.5f * ( 1.0f + ( norm.x*matWV.m00 +
										 norm.y*matWV.m10 +
										 norm.z*matWV.m20 ) ))/* * m_fTexScaleX* /;

				t.y = (0.5f * ( 1.0f - ( norm.x*matWV.m01 +
										 norm.y*matWV.m11 +
										 norm.z*matWV.m21 ) ))/* * m_fTexScaleY * /;
			}
			else
			{
				t.x = (float)vert.t.x * m_fTexScaleX;
				t.y = (float)vert.t.y * m_fTexScaleY;
			}
		}*/

		nOff += 5;
	}

}

//*****************************************************************************
//
//*****************************************************************************
void D3DRender::SetAddressU( u32 index, DWORD dwMode )
{
	g_pD3DDev->SetTextureStageState( index, D3DTSS_ADDRESSU, dwMode );
}

//*****************************************************************************
//
//*****************************************************************************
void D3DRender::SetAddressV( u32 index, DWORD dwMode )
{
	g_pD3DDev->SetTextureStageState( index, D3DTSS_ADDRESSV, dwMode );
}

//*****************************************************************************
//
//*****************************************************************************
void D3DRender::SetTextureFactor(DWORD dwCol)
{
	if (m_dwTextureFactor != dwCol)
	{
		m_dwTextureFactor = dwCol;
		g_pD3DDev->SetRenderState(D3DRS_TEXTUREFACTOR, dwCol);
	}
}



//*****************************************************************************
//
//*****************************************************************************
void D3DRender::SetTexture( u32 index,
							LPDIRECT3DTEXTURE8 lpsTexture,
							LONG  nTileLeft, LONG nTileTop,
							DWORD dwTileWidth, DWORD dwTileHeight )
{
	//lpsTexture = NULL; - For wireframe
	if (m_lpsTexture[index] != lpsTexture)
	{
		// Dereference the previous texture
		if ( m_lpsTexture[index] )
		{
			m_lpsTexture[index]->Release();
			m_lpsTexture[index] = NULL;
		}

		m_lpsTexture[index] = lpsTexture;

		if ( m_lpsTexture[index] )
		{
			HRESULT hr;
			D3DSURFACE_DESC desc;
			
			m_lpsTexture[index]->AddRef();

			m_fTileWidth[index] = (float)dwTileWidth;
			m_fTileHeight[index] = (float)dwTileHeight;

			// Update the bogus scale parameters
			m_fTexLeft[index] = (float)nTileLeft;
			m_fTexTop[index]  = (float)nTileTop;

			hr = m_lpsTexture[index]->GetLevelDesc(0, &desc);
			if (FAILED(hr))
			{
				m_fTexWidth[index] = m_fTileWidth[index];
				m_fTexHeight[index] = m_fTileHeight[index];
			}
			else
			{
				m_fTexWidth[index] = (float)desc.Width;
				m_fTexHeight[index] = (float)desc.Height;
			}

			m_fInvTileWidth[index] = 1.0f / m_fTileWidth[index];
			m_fInvTileHeight[index] = 1.0f / m_fTileHeight[index];
			m_fInvTexWidth[index] = 1.0f / m_fTexWidth[index];
			m_fInvTexHeight[index] = 1.0f / m_fTexHeight[index];

		}


		gTextureChanged = true;
	}
}

//*****************************************************************************
//
//*****************************************************************************
// Generate texture coords?
void D3DRender::SetTextureGen(BOOL bTextureGen)
{
	m_bTextureGen = bTextureGen;
}

//*****************************************************************************
//
//*****************************************************************************
void D3DRender::SetTextureScale(float fScaleX, float fScaleY)
{
	m_fTexScaleX = fScaleX;
	m_fTexScaleY = fScaleY;
}

//*****************************************************************************
//
//*****************************************************************************
void D3DRender::SetProjection(const D3DXMATRIX & mat, BOOL bPush, LONG nLoadReplace) 
{
	if (bPush)
	{
		if (m_dwProjectionTop >= (DAEDALUS_MATRIX_STACK-1))
			DBGConsole_Msg(0, "Pushing past proj stack limits!");
		else
			m_dwProjectionTop++;

		if (nLoadReplace == D3DRENDER_LOAD)
			// Load projection matrix
			m_mProjection[m_dwProjectionTop] = mat;
		else
			m_mProjection[m_dwProjectionTop] = mat * m_mProjection[m_dwProjectionTop-1];
		
	}
	else
	{
		if (nLoadReplace == D3DRENDER_LOAD)
			// Load projection matrix
			m_mProjection[m_dwProjectionTop] = mat;
		else
			m_mProjection[m_dwProjectionTop] = mat * m_mProjection[m_dwProjectionTop];

	}

}

//*****************************************************************************
//
//*****************************************************************************
void D3DRender::SetWorldView(const D3DXMATRIX & mat, BOOL bPush, LONG nLoadReplace)
{

	// ModelView
	if (bPush)
	{
		if (m_dwModelViewTop >= (DAEDALUS_MATRIX_STACK-1))
			DBGConsole_Msg(0, "Pushing past modelview stack limits!");
		else
			m_dwModelViewTop++;

		// We should store the current projection matrix...
		if (nLoadReplace == D3DRENDER_LOAD)
		{
			// Load projection matrix
			m_mModelView[m_dwModelViewTop] = mat;
		}
		else			// Multiply projection matrix
		{
			m_mModelView[m_dwModelViewTop] = mat * m_mModelView[m_dwModelViewTop-1];
		}
	}
	else	// NoPush
	{
		if (nLoadReplace == D3DRENDER_LOAD)
		{
			// Load projection matrix
			m_mModelView[m_dwModelViewTop] = mat;
		}
		else
		{
			// Multiply projection matrix
			m_mModelView[m_dwModelViewTop] = mat * m_mModelView[m_dwModelViewTop];
		}

	}
}

//*****************************************************************************
//
//*****************************************************************************
void D3DRender::PopProjection()
{
	if (m_dwProjectionTop > 0)
		m_dwProjectionTop--;
	//else
	//	DBGConsole_Msg(0, "Popping past projection stack limits");
}

//*****************************************************************************
//
//*****************************************************************************
void D3DRender::PopWorldView()
{
	if (m_dwModelViewTop > 0)
		m_dwModelViewTop--;
	//else
	//	DBGConsole_Msg(0, "Popping past worldview stack limits");
}





#ifdef VRML_EXPORT_SUPPORT
//*****************************************************************************
//
//*****************************************************************************
void EndSceneDump()
{
	m_DumpingScene = false;

	FILE * fh = fopen( "c:\\temp\\daedalus.wrl", "w" );
	if ( fh == NULL )
		return;

	fprintf( fh, "#VRML V2.0 utf8\n" );
	fprintf( fh, "\n" );
	fprintf( fh, "# Produced by Daedalus\n" );
	fprintf( fh, "\n" );
	fprintf( fh, "" );

	for ( u32 o = 0; o < m_arrObjects.size(); o++ )
	{
		const CHAR * name = "MAGGIE";

		u32 vert_count = m_arrObjects[ o ].GetNumVerts();
		
		fprintf( fh, "DEF %s Transform {\n", name );
		fprintf( fh, "  translation 0 0 0\n" );
		fprintf( fh, "  children [\n" );
		fprintf( fh, "    Shape {\n" );
		fprintf( fh, "      appearance Appearance {\n" );
		fprintf( fh, "        material Material {\n" );
		fprintf( fh, "          diffuseColor 1 1 0\n" );
		fprintf( fh, "          ambientIntensity 0.6667\n" );
		fprintf( fh, "          specularColor 0.7981 0.7981 0.7944\n" );
		fprintf( fh, "          shininess 0.8195\n" );
		fprintf( fh, "          transparency 0\n" );
		fprintf( fh, "        }\n" );

		if ( strlen( m_arrObjects[ o ].GetTextureURL() ) > 0 )
		{
			fprintf( fh, "        texture ImageTexture {\n" );
			fprintf( fh, "          url \"%s\"\n", m_arrObjects[ o ].GetTextureURL() );
			fprintf( fh, "        }\n" );
		}
		fprintf( fh, "      }\n" );

      	fprintf( fh, "      geometry DEF %s-FACES IndexedFaceSet {\n", name );
		fprintf( fh, "        ccw TRUE\n" );
		fprintf( fh, "        solid FALSE\n" );
		fprintf( fh, "        colorPerVertex TRUE\n" );
		fprintf( fh, "        color Color { color [\n" );
		//fprintf( fh, "          1 1 0, 0 0 1, 0 0 0, 1 1 1, 0.6667 0 0\n);

		for ( u32 i = 0; i < vert_count; i++ )
		{
			DWORD col = m_arrObjects[ o ].GetColor( i );
			float r = RGBA_GETRED( col ) / 255.0f;
			float g = RGBA_GETGREEN( col ) / 255.0f;
			float b = RGBA_GETBLUE( col ) / 255.0f;
			fprintf( fh, " %f %f %f,\n", r, g, b );
		}

		fprintf( fh, " ] }\n" );



		//
		fprintf( fh, "        coord DEF %s-COORD Coordinate { point [\n", name );
		for ( i = 0; i < vert_count; i++ )
		{
			D3DXVECTOR3 pos = m_arrObjects[ o ].GetPos( i );
			fprintf( fh, " %f %f %f,\n", pos.x, pos.y, pos.z );
		}
		fprintf( fh, " ] }\n" );

		//
        fprintf( fh, "        texCoord DEF %s-TEXCOORD TextureCoordinate { point [\n", name );
		for ( i = 0; i < vert_count; i++ )
		{
			D3DXVECTOR2 uv = m_arrObjects[ o ].GetTextureCoords( i );
			fprintf( fh, " %f %f,\n", uv.x, uv.y );
		}
		fprintf( fh, " ] }\n" );

		//
		fprintf( fh, "        coordIndex [\n" );
		for ( i = 0; i < vert_count; i+=3 )
		{
			fprintf( fh, " %d, %d, %d, %d,\n", i+0, i+1, i+2, -1 );
		}
		fprintf( fh, " ]\n" );


		fprintf( fh, "}\n" );
		fprintf( fh, "}\n" );
		fprintf( fh, " ]\n" );
		fprintf( fh, "}\n" );

	}


	fclose(fh);
}
#endif // VRML_EXPORT_SUPPORT


#ifdef USE_PIXEL_SHADERS
//*************************************************************************************
// 
// Pixel shader code
//
//*************************************************************************************

//*************************************************************************************
// 
//*************************************************************************************
void	D3DRender::InitCycle1PixelShader()
{
	// Censored!
}

//*************************************************************************************
// 
//*************************************************************************************
void	D3DRender::InitCycle2PixelShader()
{
	// Censored!
}

//*************************************************************************************
// 
//*************************************************************************************
void	D3DRender::InitCycle2FogShadePixelShader()
{
	// Censored!
}

#endif //USE_PIXEL_SHADERS


// Draw nodes on vertices
/*D3DXVECTOR3 a, b, c;

a.x = (v0.x * 160) + 160; a.y = (-v0.y * 120) + 120;
b.x = (v1.x * 160) + 160; b.y = (-v1.y * 120) + 120;
c.x = (v2.x * 160) + 160; c.y = (-v2.y * 120) + 120;

g_Renderer->FillRect(a.x, a.y, a.x+2, a.y+2, RGBA_MAKE(255,0,255,255));
g_Renderer->FillRect(b.x, b.y, b.x+2, b.y+2, RGBA_MAKE(255,0,255,255));
g_Renderer->FillRect(c.x, c.y, c.x+2, c.y+2, RGBA_MAKE(255,0,255,255));* /
*/


