/*
Copyright (C) 2001 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
	  
*/

// Manages textures for RDP code
// Uses a HashTable (hashing on TImg) to allow quick access
//  to previously used textures

#include "stdafx.h"

#include "TextureCache.h"
#include "ConvertImage.h"
#include "RDP.h"

#include "OSHLE\ultra_gbi.h"		// For image size/formats
#include "Utility\Profiler.h"
#include "Core\ROM.h"
#include "Core\Memory.h"

#include "Debug\DBGConsole.h"

// Returns the first prime greater than or equal to nFirst
static u32 GetNextPrime(u32 nFirst)
{
	u32 nCurrent;
	
	u32 i;
	
	nCurrent = nFirst;
	
	// Just make sure it's odd
	if ((nCurrent % 2) == 0)
		nCurrent++;
	
	for (;;)
	{
		u32 nSqrtCurrent;
		bool bIsComposite;
		
		// nSqrtCurrent = nCurrent^0.5 + 1 (round up)
		nSqrtCurrent = (u32)sqrtf((f32)nCurrent) + 1;
		
		bIsComposite = false;
		
		// Test all odd numbers from 3..nSqrtCurrent
		for (i = 3; i <= nSqrtCurrent; i+=2)
		{
			if ((nCurrent % i) == 0)
			{
				bIsComposite = true;
				break;
			}
		}
		
		if (!bIsComposite)
		{			
			return nCurrent;
		}
		
		// Select next odd candidate...
		nCurrent += 2;
	}
	
}


// Implementation

typedef struct TextureEntry
{
	TextureEntry() : pTexture(NULL), pMirroredTexture(NULL) {}

	~TextureEntry()
	{
		if (pTexture)
		{
			delete pTexture;
			pTexture = NULL;
		}
		if (pMirroredTexture)
		{
			delete pMirroredTexture;
			pMirroredTexture = NULL;
		}
	}

	//
	// Data
	//
	struct TextureEntry * pNext;

	TextureInfo ti;
	u32 dwCRC;

	u32 dwUses;			// Total times used (for stats)
	u32 FrameLastUsed;	// Frame # that this was last used

	CTexture	* pTexture;
	// TODO: when nXTimes and nYTimes in the mirror emulator are correctly determined, these should be put here
	CTexture * pMirroredTexture;

} TextureEntry;



//*****************************************************************************
// Texture cache implementation
//*****************************************************************************
class ITextureCache : public CTextureCache
{
protected:
	TextureEntry * CreateEntry(TextureInfo * pti);
	void AddTextureEntry(TextureEntry *pEntry);
	void RemoveTextureEntry(TextureEntry * pEntry);
	
	void AddToRecycleList(TextureEntry *pEntry);
	TextureEntry * ReviveUsedTexture( u32 width, u32 height );
	
	
	TextureEntry * GetEntry(TextureInfo * pti);
	TextureEntry * GetEntryAndUpdate(TextureInfo * pti);
	
	u32 Hash(u32 dwValue);
	
	void DecompressTexture(TextureEntry * pEntry);
	void DecompressTmemTexture(TextureEntry * pEntry);

	u32 CalculateCRC(TextureInfo * pti) const;
	
protected:
	TextureEntry * m_pFirstUsedSurface;
	TextureEntry ** m_pTextureHash;
	u32 m_dwTextureHashSize;
	BOOL m_bDumpTextures;
	
public:
	ITextureCache();
	~ITextureCache();
	
	void PurgeOldTextures();
	void DropTextures();

	void SetDumpTextures( BOOL bDumpTextures ) { m_bDumpTextures = bDumpTextures; }
	BOOL GetDumpTextures( ) const { return m_bDumpTextures; }


	HRESULT InvalidateDeviceObjects();
	
#ifdef MIRRORED_TEXTURE_SUPPORT
	CTexture * GetTexture(TextureInfo * pti, BOOL bMirrored);
#else
	CTexture * GetTexture( TextureInfo * pti );
#endif
	
	HRESULT Initialise(u32 size);
};


// Interface

//*****************************************************************************
//
//*****************************************************************************

HRESULT CUnique< CTextureCache >::Create()
{
	HRESULT hr;
	
	ITextureCache * pNewInstance;
	
	_ASSERTE(m_pInstance == NULL);
	
	pNewInstance = new ITextureCache();
	if (!pNewInstance)
	{
		hr = E_OUTOFMEMORY;
	}
	else
	{
		hr = pNewInstance->Initialise(8000);
		if (FAILED(hr))
		{
			delete pNewInstance;
			pNewInstance = NULL;
		}
	}
	
	m_pInstance = pNewInstance;
	
	return hr;
}


///////////////////////////////////////////////////////////////////////
//
///////////////////////////////////////////////////////////////////////

ITextureCache::ITextureCache() :
	m_pFirstUsedSurface(NULL),
	m_pTextureHash(NULL),
	m_dwTextureHashSize(0),
	m_bDumpTextures(FALSE)
{
}

ITextureCache::~ITextureCache()
{
	DropTextures();
	
	while (m_pFirstUsedSurface)
	{
		TextureEntry * pVictim = m_pFirstUsedSurface;
		m_pFirstUsedSurface = pVictim->pNext;
		
		delete pVictim;
	}

	delete []m_pTextureHash;
	m_pTextureHash = NULL;	
}

//
// Called when the device-dependent objects are about to be lost.
//
HRESULT ITextureCache::InvalidateDeviceObjects()
{
	DropTextures();
	
	while (m_pFirstUsedSurface)
	{
		TextureEntry * pVictim = m_pFirstUsedSurface;
		m_pFirstUsedSurface = pVictim->pNext;
		
		delete pVictim;
	}

	return S_OK;
}



HRESULT ITextureCache::Initialise(u32 size)
{
	size = GetNextPrime(size);
	
	m_pTextureHash = new TextureEntry *[size];
	if (m_pTextureHash == NULL)
		return E_OUTOFMEMORY;
	
	m_dwTextureHashSize = size;
	
	for (u32 i = 0; i < size; i++)
		m_pTextureHash[i] = NULL;
	
	return S_OK;	// Success!
	
}

// Purge any textures whos last usage was over 5 seconds ago
void ITextureCache::PurgeOldTextures()
{
	if (m_pTextureHash == NULL)
		return;
	
	Lock();
	
	static const u32 dwFramesToKill = 5*30;			// 5 secs at 30 fps
	static const u32 dwFramesToDelete = 30*30;		// 30 secs at 30 fps
	
	for ( u32 i = 0; i < m_dwTextureHashSize; i++ )
	{
		TextureEntry * pEntry;
		TextureEntry * pNext;
		
		pEntry = m_pTextureHash[i];
		while (pEntry)
		{
			pNext = pEntry->pNext;
			
			if ( gRDPFrame - pEntry->FrameLastUsed > dwFramesToKill )
			{
				RemoveTextureEntry(pEntry);
			}
			pEntry = pNext;
		}
	}
	
	
	// Remove any old textures that haven't been recycled in 1 minute or so
	// Normally these would be reused
	TextureEntry * pPrev;
	TextureEntry * pCurr;
	TextureEntry * pNext;
	
	
	pPrev = NULL;
	pCurr = m_pFirstUsedSurface;
	
	while (pCurr)
	{
		pNext = pCurr->pNext;
		
		if ( gRDPFrame - pCurr->FrameLastUsed > dwFramesToDelete )
		{
			// Everything from this point on should be too old!
			// Remove from list
			if (pPrev != NULL) pPrev->pNext        = pCurr->pNext;
			else			   m_pFirstUsedSurface = pCurr->pNext;
			
			//DBGConsole_Msg(0, "Killing old used texture (%d x %d)", pCurr->dwWidth, pCurr->dwHeight);
			
			delete pCurr;
			
			// pPrev remains the same
			pCurr = pNext;	
		}
		else
		{
			pPrev = pCurr;
			pCurr = pNext;
		}
	}
	
	Unlock();
}

void ITextureCache::DropTextures()
{
	if (m_pTextureHash == NULL)
		return;
	
	Lock();
	
	u32 dwCount = 0;
	u32 dwTotalUses = 0;
	
	for (u32 i = 0; i < m_dwTextureHashSize; i++)
	{
		while (m_pTextureHash[i])
		{
			TextureEntry *pTVictim = m_pTextureHash[i];
			m_pTextureHash[i] = pTVictim->pNext;
			
			dwTotalUses += pTVictim->dwUses;
			dwCount++;
			
			AddToRecycleList(pTVictim);
		}
	}
	
	DBGConsole_Msg(0, "Texture Handler: %d entries in hash table, %d uses.", dwCount, dwTotalUses);
	
	Unlock();
	
}

// Add to the recycle list
void ITextureCache::AddToRecycleList(TextureEntry *pEntry)
{
	if (pEntry->pTexture == NULL)
	{
		// No point in saving!
		delete pEntry;
	}
	else
	{
		// Add to the list
		pEntry->pNext = m_pFirstUsedSurface;
		m_pFirstUsedSurface = pEntry;
	}
}

// Search for a texture of the specified dimensions to recycle
TextureEntry * ITextureCache::ReviveUsedTexture( u32 width, u32 height )
{
	TextureEntry * pPrev;
	TextureEntry * pCurr;
	
	pPrev = NULL;
	pCurr = m_pFirstUsedSurface;
	
	while (pCurr)
	{
		if (pCurr->ti.Width == width &&
			pCurr->ti.Height == height)
		{
			// Remove from list
			if (pPrev != NULL) pPrev->pNext        = pCurr->pNext;
			else			   m_pFirstUsedSurface = pCurr->pNext;
			
			//DBGConsole_Msg(0, "Reviving used texture (%d x %d)", dwWidth, dwHeight);
			// Initialise any fields:
			return pCurr;
		}
		
		pPrev = pCurr;
		pCurr = pCurr->pNext;
	}
	
	
	
	return NULL;
}



u32 ITextureCache::Hash(u32 dwValue)
{
	// Divide by four, because most textures will be on a 4 byte boundry, so bottom four
	// bits are null
	return (dwValue>>2) % m_dwTextureHashSize;
}


void ITextureCache::AddTextureEntry(TextureEntry *pEntry)
{	
	u32 dwKey = Hash(pEntry->ti.Address);
	
	if (m_pTextureHash == NULL)
		return;
		
	Lock();
	
	TextureEntry **p = &m_pTextureHash[dwKey];
	
	// Add to head (not tail, for speed - new textures are more likely to be accessed next)
	pEntry->pNext = m_pTextureHash[dwKey];
	m_pTextureHash[dwKey] = pEntry;
	
	Unlock();
	
}

TextureEntry * ITextureCache::GetEntry(TextureInfo * pti)
{
	TextureEntry *pEntry;
	
	if (m_pTextureHash == NULL)
		return NULL;
	
	// See if it is already in the hash table
	u32 dwKey = Hash(pti->Address);
	for (pEntry = m_pTextureHash[dwKey]; pEntry; pEntry = pEntry->pNext)
	{
		if ( pEntry->ti.IsEqual(*pti) )
			return pEntry;
	}
	return NULL;
}

void ITextureCache::RemoveTextureEntry(TextureEntry * pEntry)
{
	TextureEntry * pPrev;
	TextureEntry * pCurr;
	
	if (m_pTextureHash == NULL)
		return;
	
	Lock();
	
	//DBGConsole_Msg(0, "Remove Texture entry!");
	
	// See if it is already in the hash table
	u32 dwKey = Hash(pEntry->ti.Address);
	
	pPrev = NULL;
	pCurr = m_pTextureHash[dwKey];
	
	while (pCurr)
	{
		// Check that the attributes match
		if (pCurr->ti.IsEqual(pEntry->ti))
		{
			if (pPrev != NULL) pPrev->pNext = pCurr->pNext;
			else			   m_pTextureHash[dwKey] = pCurr->pNext;

			break;
		}
		
		pPrev = pCurr;
		pCurr = pCurr->pNext;
	}
	
	Unlock();
	
	if (pCurr == NULL)
	{
		DBGConsole_Msg(0, "Entry not found!!!");
	}
	
	AddToRecycleList(pEntry);
}

TextureEntry * ITextureCache::CreateEntry(TextureInfo * pti)
{
	TextureEntry * pEntry;
	
	// Find a used texture
	pEntry = ReviveUsedTexture( pti->Width, pti->Height );
	if (pEntry == NULL)
	{
		// Couldn't find on - recreate!
		pEntry = new TextureEntry;
		if (pEntry == NULL)
			return NULL;
		
		pEntry->pTexture = CTexture::Create( pti->Width, pti->Height );

		// Ignore failure for now. This prevents us trying to recreate
		// the texture each frame, which would really slow things down
	}
	
	// Initialise
	pEntry->pNext = NULL;
	pEntry->ti = *pti;
	pEntry->dwUses = 0;
	pEntry->FrameLastUsed = gRDPFrame;
	pEntry->dwCRC = CalculateCRC(pti);
	
	// Add to the hash table
	AddTextureEntry(pEntry);
	return pEntry;	
}

u32 ITextureCache::CalculateCRC(TextureInfo * pti) const
{
	u32 dwCRC;
	u8 * pStart;
	u32 dwBytesPerLine;
	//s32 x,y;

	// If CRC checking is disabled, alwasy return 0
	if ( g_ROM.settings.bDisableTextureCRC )
		return 0;
	
	dwBytesPerLine = ((pti->Width<<pti->Size)+1)/2;

	//DBGConsole_Msg(0, "BytesPerLine: %d", dwBytesPerLine);
	
	// A very simple crc - just summation
	dwCRC = 0;

	// The assembly below replaces and improves on this
	/*pStart = g_pu8RamBase + pti->dwAddress;
	pStart += (pti->nTop * pti->dwPitch) + (((pti->nLeft<<pti->dwSize)+1)/2);
	for (y = 0; y < pti->dwHeight; y++)		// Do every nth line?
	{
		// Byte fiddling won't work, but this probably doesn't matter
		// Now process 4 bytes at a time
		for (x = 0; x < dwBytesPerLine; x+=4)
		{
			dwCRC = (dwCRC + *(u32*)&pStart[x]);
		}

		pStart += pti->dwPitch;
	}

	u32 dwTestCRC = dwCRC;*/

	pStart = g_pu8RamBase + pti->Address;
	pStart += (pti->Top * pti->Pitch) + (((pti->Left<<pti->Size)+1)/2);

	u32 ptiHeight = pti->Height;
	u32 ptiPitch = pti->Pitch;
	__asm
	{
		xor		eax, eax			// Clear checksum

		mov		ebx, ptiPitch
		mov		ecx, ptiHeight

		mov		esi, dword ptr [pStart]

		// Calculate initial x counter value
		mov		edx, dwBytesPerLine
		add		edx, 3				// Add 3 to round up result
		shr		edx, 2
		jz		done				// Check for BytesPerLine <= 4

		// Check which version of the code to use: optimise
		// for cases where dwBytesPerLine is a multiple of 16 bytes
		test	edx, 3
		jz		do_mult_of_4

	// The x counter is not a multiple of 4, so some extra code
	// is needed to bring it down to the boundary
	do_not_mult_of_4:
		mov		edi, edx
	x_loop_top_no4:
		test	edi, 3
		jz		do4
	//do1:
		add		eax, [esi+edi*4-4]
		dec		edi
		jnz		x_loop_top_no4
		jmp		done_x_no4
	do4:
		add		eax, [esi+edi*4-4]
		add		eax, [esi+edi*4-8]
		add		eax, [esi+edi*4-12]
		add		eax, [esi+edi*4-16]
		sub		edi, 4
		jnz		do4							// still more to do
	done_x_no4:
		add		esi, ebx					// add pitch to base pointer

		loop	do_not_mult_of_4			// dec ecx, jump to y_loop_top if not zero
		jmp		done


	// The x counter is known to be a multiple of 4, so no code
	// is needed to do the surplus bits
	do_mult_of_4:
		mov		edi, edx
	x_loop_top_4:
		add		eax, [esi+edi*4-4]
		add		eax, [esi+edi*4-8]
		add		eax, [esi+edi*4-12]
		add		eax, [esi+edi*4-16]
		sub		edi, 4
		jnz		x_loop_top_4			// still more to do
	//done_x_4:
		add		esi, ebx				// add pitch to base pointer
		loop	do_mult_of_4			// dec ecx, jump to y_loop_top if not zero
	done:
		mov		dwCRC, eax
	}

	/*if ( dwTestCRC != dwCRC )
	{
		DBGConsole_Msg(0, "CRCs don't match");
	}*/

	if (pti->Format == G_IM_FMT_CI)
	{
		u32 bytes;
		if ( pti->Size == G_IM_SIZ_4b )		bytes = 16  * 4;
		else								bytes = 256 * 4;

		pStart = &gTextureMemory[ pti->TmemPalAddress << 3 ];
		for ( u32 x = 0; x < bytes; x+=4)
		{
			dwCRC = (dwCRC + *(u32*)&pStart[x]);
		}
	}

	return dwCRC;
}


// Lkb Mirrored texture support:

static void MirrorEmulator_DrawLine(DrawInfo& destInfo, DrawInfo& srcInfo, u32 * pSource, u32 * pDest, u32 nWidth, BOOL bFlipLeftRight)
{
	if(!bFlipLeftRight)
	{
		memcpy(pDest, pSource, nWidth * 4);
	}
	else
	{
		u32 * pMaxDest = pDest + nWidth;
		pSource += nWidth - 1;
		for(; pDest < pMaxDest; pDest++, pSource--)
		{
			*pDest = *pSource;
		}
	}
}

static void MirrorEmulator_Draw(DrawInfo& destInfo, DrawInfo& srcInfo, u32 nDestX, u32 nDestY, BOOL bFlipLeftRight, BOOL bFlipUpDown)
{
	LPBYTE pDest = (LPBYTE)((u32)destInfo.Surface + (destInfo.Pitch * nDestY) + (4 * nDestX));
	LPBYTE pMaxDest = pDest + (destInfo.Pitch * srcInfo.Height);
	LPBYTE pSource = (LPBYTE)(srcInfo.Surface);
	if(!bFlipUpDown)
	{
		for(; pDest < pMaxDest; pDest += destInfo.Pitch, pSource += srcInfo.Pitch)
		{
			MirrorEmulator_DrawLine(destInfo, srcInfo, (u32 *)pSource, (u32 *)pDest, srcInfo.Width, bFlipLeftRight);
		}
	}
	else
	{
		pSource += (srcInfo.Pitch * (srcInfo.Height - 1));
		for(; pDest < pMaxDest; pDest += destInfo.Pitch, pSource -= srcInfo.Pitch)
		{
			MirrorEmulator_DrawLine(destInfo, srcInfo, (u32 *)pSource, (u32 *)pDest, srcInfo.Width, bFlipLeftRight);
		}
	}
}



static void CreateMirroredTexture(TextureEntry *pEntry)
{
	if(pEntry->pMirroredTexture)
	{
		return;
	}

	// FIXME: Compute the correct values. 2/2 seems to always work correctly in Mario64
	u32 nXTimes = 2;
	u32 nYTimes = 2;
	
	DrawInfo srcInfo;		
	pEntry->pTexture->StartUpdate(&srcInfo);
	u32 nWidth = srcInfo.Width;
	u32 nHeight = srcInfo.Height;

	pEntry->pMirroredTexture = CTexture::Create(nWidth * nXTimes, nHeight * nYTimes);
	DrawInfo destInfo;
	pEntry->pMirroredTexture->StartUpdate(&destInfo);
	
	
	for(u32 nY = 0; nY < nYTimes; nY++)
	{
		for(u32 nX = 0; nX < nXTimes; nX++)
		{
			MirrorEmulator_Draw(destInfo, srcInfo, nWidth * nX, nHeight * nY, nX & 0x1, nY & 0x1);
		}
	}

	pEntry->pMirroredTexture->EndUpdate(&destInfo);
	pEntry->pTexture->EndUpdate(&srcInfo);
}

#ifdef MIRRORED_TEXTURE_SUPPORT

CTexture * ITextureCache::GetTexture(TextureInfo * pti, BOOL bMirrored)
{
	TextureEntry * pEntry;
	
	START_PROFILE( CProfiler::PROFILE_GFX_TCACHE );
		pEntry = GetEntryAndUpdate(pti);
	STOP_PROFILE( CProfiler::PROFILE_GFX_TCACHE );
	if (pEntry == NULL)
		return NULL;

	if (!bMirrored)
	{
		return pEntry->pTexture;
	}
	else
	{
		// Create if neccessary
		if (!pEntry->pMirroredTexture)
		{
			CreateMirroredTexture(pEntry);
		}
		return pEntry->pMirroredTexture;
	}
}
#else

CTexture * ITextureCache::GetTexture(TextureInfo * pti)
{
	TextureEntry * pEntry;
	
	START_PROFILE( CProfiler::PROFILE_GFX_TCACHE );
		pEntry = GetEntryAndUpdate(pti);
	STOP_PROFILE( CProfiler::PROFILE_GFX_TCACHE );
	if (pEntry == NULL)
		return NULL;

	return pEntry->pTexture;
}

#endif

// If already in table, return cached copy
// Otherwise, create surfaces, and load texture into memory
TextureEntry * ITextureCache::GetEntryAndUpdate(TextureInfo * pti)
{
	TextureEntry *pEntry;
	u32 dwCRC;
	
	//
	// Retrieve the texture from the cache (if it already exists)
	//
	pEntry = GetEntry(pti);
	if (pEntry == NULL)
	{
		// We need to create a new entry, and add it to the hash table.
		pEntry = CreateEntry(pti);
		if (pEntry == NULL)
			return NULL;

	}
	else
	{
		// See if we've already checked that this entry is valid this frame
		if (pEntry->FrameLastUsed == gRDPFrame)
		{
			pEntry->dwUses++;
			return pEntry;
		}

		//
		// Generate and check the current crc
		//
		dwCRC = CalculateCRC(pti);

		// Check if the CRCs match, if so return. Don't do this for newly created entries
		if (pEntry->dwCRC == dwCRC)
		{
			// Tile is ok, return
			pEntry->dwUses++;
			pEntry->FrameLastUsed = gRDPFrame;
			return pEntry;
		}

		// The crc needs updating
		pEntry->dwCRC = dwCRC;
	}

		
	//
	// The texture needs to be updated for this frame
	//
	START_PROFILE( CProfiler::PROFILE_GFX_CONVERT );
	try 
	{
		if (pEntry->pTexture != NULL)
		{
#if RDP_EMULATE_TMEM
			DecompressTmemTexture(pEntry);		
#else
			DecompressTexture(pEntry);		
#endif
		}
	}
	catch (...)
	{
		DBGConsole_Msg(0, "Exception in texture decompression");
		if (!g_bTrapExceptions)
			throw;
	}
	STOP_PROFILE( CProfiler::PROFILE_GFX_CONVERT );
	
	return pEntry;
	
}



//*************************************************************************************
// 
//*************************************************************************************

static const char *pszImgFormat[8] = {"RGBA", "YUV", "CI", "IA", "I", "?1", "?2", "?3"};
static const u8 pnImgSize[4]   = {4, 8, 16, 32};

void ITextureCache::DecompressTexture(TextureEntry * pEntry)
{
	static u32 dwCount = 0;
	
	ConvertFunction pF = gConvertFunctions[ pEntry->ti.Format ][ pEntry->ti.Size ];

	if ( pF )
	{
		pF( pEntry->pTexture, pEntry->ti );

#ifdef LIBPNG_SUPPORT
		if ( m_bDumpTextures )
		{
			
			TCHAR szFileName[MAX_PATH+1];
			TCHAR szFilePath[MAX_PATH+1];
			TCHAR szDumpDir[MAX_PATH+1];
			
			wsprintf(szDumpDir, "%s\\%s", TEXT("tile_dumps"), g_ROM.settings.szGameName);
			
			Dump_GetDumpDirectory(szFilePath, szDumpDir);
			
			wsprintf(szFileName, TEXT("%08.8x-%s_%dbpp-%dx%d-%dx%d.png"),
				pEntry->ti.Address, pszImgFormat[pEntry->ti.Format], pnImgSize[pEntry->ti.Size],
				pEntry->ti.Left, pEntry->ti.Top,
				pEntry->ti.Width,pEntry->ti.Height);
			
			PathAppend(szFilePath, szFileName);
			
			pEntry->pTexture->DumpImageAsPNG(szFilePath, pEntry->ti.Width,pEntry->ti.Height);
		}
#endif		
	}
	else
	{
		static bool warning = false;
		
		if ( !warning )
		{			
			DBGConsole_Msg(0, "DecompressTexture: Unable to decompress %s/%dbpp", pszImgFormat[pEntry->ti.Format], pnImgSize[pEntry->ti.Size]);	
			//warning = true;
		}
	}
	
	dwCount++;	
}

//*************************************************************************************
// Load a texture from tmem
//*************************************************************************************
void ITextureCache::DecompressTmemTexture(TextureEntry * pEntry)
{
	static u32 dwCount = 0;

	ConvertFunction pF = gTmemConvertFunctions[ pEntry->ti.Format ][ pEntry->ti.Size ];
	
	if ( pF )
	{
		pF( pEntry->pTexture, pEntry->ti );

#ifdef LIBPNG_SUPPORT
		if ( m_bDumpTextures )
		{
			
			TCHAR szFileName[MAX_PATH+1];
			TCHAR szFilePath[MAX_PATH+1];
			TCHAR szDumpDir[MAX_PATH+1];
			
			wsprintf(szDumpDir, "%s\\%s", TEXT("tile_dumps"), g_ROM.settings.szGameName);
			
			Dump_GetDumpDirectory(szFilePath, szDumpDir);
			
			wsprintf(szFileName, TEXT("%08.8x-%s_%dbpp-%dx%d-%dx%d.png"),
				pEntry->ti.Address, pszImgFormat[pEntry->ti.Format], pnImgSize[pEntry->ti.Size],
				pEntry->ti.Left, pEntry->ti.Top,
				pEntry->ti.Width,pEntry->ti.Height);
			
			PathAppend(szFilePath, szFileName);
			
			pEntry->pTexture->DumpImageAsPNG(szFilePath, pEntry->ti.Width,pEntry->ti.Height);
		}
#endif	
		
	}
	else
	{
		static bool warning = false;
		
		if ( !warning )
		{			
			DBGConsole_Msg(0, "DecompressTexture: Unable to decompress %s/%dbpp", pszImgFormat[pEntry->ti.Format], pnImgSize[pEntry->ti.Size]);	
			//warning = true;
		}
	}
	
	
	dwCount++;	
}


